import json
import logging
import math
import sys
from datetime import datetime
from unittest.mock import patch

import pytest
import requests
from jira import JIRAError

from jf_ingest.config import IssueMetadata
from jf_ingest.utils import (
    RETRY_EXPONENT_BASE,
    RetryLimitExceeded,
    batch_iterable,
    batch_iterable_by_bytes_size,
    format_date_to_jql,
    get_object_bytes_size,
    get_wait_time,
    hash_filename,
    retry_for_status,
)

logger = logging.getLogger(__name__)

MOCK_RETRY_AFTER_TIME = 12345


class MockJiraErrorResponseWithRetryAfterAsInt:
    headers = {"Retry-After": MOCK_RETRY_AFTER_TIME}


class MockJiraErrorResponseWithRetryAfterAsStr:
    headers = {"Retry-After": str(MOCK_RETRY_AFTER_TIME)}


class MockJiraErrorResponseWithZeroRetryAfter:
    headers = {"Retry-After": 0}


class MockJiraErrorResponseWithNonIntRetryAfter:
    headers = {"Retry-After": "CAT"}


MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT = JIRAError(
    status_code=429,
    text="This is a spoofed Jira Error",
    response=MockJiraErrorResponseWithRetryAfterAsInt(),
)
MOCKED_504_JIRA_ERROR_WITH_RETRY_AS_INT = JIRAError(
    status_code=504,
    text="This is a spoofed Jira Error",
    response=MockJiraErrorResponseWithRetryAfterAsInt(),
)
MOCKED_404_JIRA_ERROR_WITH_RETRY_AS_INT = JIRAError(
    status_code=404,
    text="This is a spoofed Jira Error",
    response=MockJiraErrorResponseWithRetryAfterAsInt(),
)
MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_STR = JIRAError(
    status_code=429,
    text="This is a spoofed Jira Error",
    response=MockJiraErrorResponseWithRetryAfterAsStr(),
)

MOCKED_429_JIRA_ERROR_RETRY_AFTER_IS_ZERO = JIRAError(
    status_code=429,
    text="This is a spoofed Jira Error with Retry After set to 0",
    response=MockJiraErrorResponseWithZeroRetryAfter(),
)

MOCKED_429_JIRA_ERROR_RETRY_AFTER_IS_NOT_INT = JIRAError(
    status_code=429,
    text="This is a spoofed Jira Error with Retry After set to a non int",
    response=MockJiraErrorResponseWithNonIntRetryAfter(),
)


def test_get_default_wait_time():
    # Get Default Wait Times
    for retry_num in range(0, 10):
        assert get_wait_time(e=None, retries=retry_num) == RETRY_EXPONENT_BASE**retry_num


def test_get_wait_times_for_retry_error():
    # Test when we have a valid Error
    mock_error = MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT

    for retry_num in range(0, 5):
        logger.info(f"Retry number {retry_num}...")
        assert get_wait_time(mock_error, retries=retry_num) == MOCK_RETRY_AFTER_TIME


def test_get_wait_times_for_retry_error_when_retry_is_str():
    # Test when we have a valid Error
    mock_error = MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_STR

    for retry_num in range(0, 5):
        logger.info(f"Retry number {retry_num}...")
        assert get_wait_time(mock_error, retries=retry_num) == MOCK_RETRY_AFTER_TIME


def test_get_wait_times_for_retry_error_zero():
    # Test when we have a valid Error
    mock_error = MOCKED_429_JIRA_ERROR_RETRY_AFTER_IS_ZERO

    for retry_num in range(0, 10):
        logger.info(f"Retry number {retry_num}...")
        expected_wait_time = RETRY_EXPONENT_BASE**retry_num
        # Retry times SHOULD NEVER be zero
        assert get_wait_time(mock_error, retries=retry_num) != 0
        assert get_wait_time(mock_error, retries=retry_num) == expected_wait_time


def test_get_wait_times_for_retry_error_non_zero():
    # Test when we have a valid Error
    mock_error = MOCKED_429_JIRA_ERROR_RETRY_AFTER_IS_NOT_INT

    for retry_num in range(0, 10):
        logger.info(f"Retry number {retry_num}...")
        expected_wait_time = RETRY_EXPONENT_BASE**retry_num
        # Retry times SHOULD NEVER be zero
        assert get_wait_time(mock_error, retries=retry_num) != 0
        assert get_wait_time(mock_error, retries=retry_num) == expected_wait_time


def test_get_wait_time_for_non_retry_error():
    mock_non_retry_exception = Exception()
    retries = 5
    assert (
        get_wait_time(mock_non_retry_exception, retries=retries) == RETRY_EXPONENT_BASE**retries
    )


def test_retry_for_status_timeout():
    def always_raise_429():
        raise MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT

    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        with pytest.raises(RetryLimitExceeded):
            retry_for_status(always_raise_429)

        try:
            retry_for_status(always_raise_429)
        except RetryLimitExceeded:
            m.assert_called_with(MOCK_RETRY_AFTER_TIME)


def test_retry_for_status_retry_works_429():
    arg_dict = {"retry_count": 0}
    success_message = "Success!"

    def raise_429_three_times():
        if arg_dict["retry_count"] == 3:
            return success_message
        else:
            arg_dict["retry_count"] += 1
            raise MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT

    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        assert retry_for_status(raise_429_three_times) == success_message
        m.assert_called_with(MOCK_RETRY_AFTER_TIME)


def test_retry_for_status_retry_works_504():
    arg_dict = {"retry_count": 0}
    success_message = "Success!"

    def raise_504_three_times():
        if arg_dict["retry_count"] == 3:
            return success_message
        else:
            arg_dict["retry_count"] += 1
            raise MOCKED_504_JIRA_ERROR_WITH_RETRY_AS_INT

    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        assert retry_for_status(raise_504_three_times) == success_message
        m.assert_called_with(MOCK_RETRY_AFTER_TIME)


def test_retry_for_status_retry_works_custom_error_code_list():
    def always_raise_429():
        raise MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT

    arg_dict = {"retry_count": 0}
    success_message = "Success!"

    some_error_codes = []
    some_error_codes.extend([400, 500])
    some_error_codes.append(404)

    def raise_404_three_times(test_arg: str):
        assert bool(test_arg)
        if arg_dict["retry_count"] == 3:
            return success_message
        else:
            arg_dict["retry_count"] += 1
            raise MOCKED_404_JIRA_ERROR_WITH_RETRY_AS_INT

    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        assert (
            retry_for_status(
                raise_404_three_times, 'test_input', statuses_to_retry=some_error_codes
            )
            == success_message
        )
        m.assert_called_with(MOCK_RETRY_AFTER_TIME)

    with pytest.raises(JIRAError):
        # Assert that we raise codes that are not included
        retry_for_status(always_raise_429, statuses_to_retry=some_error_codes)


def test_retry_for_status_with_generic_error():
    def raise_non_429():
        raise Exception("Generic Error Testing")

    with pytest.raises(Exception):
        retry_for_status(raise_non_429)


def test_retry_for_status_retry_works_readtimeout():
    arg_dict = {"retry_count": 0}
    success_message = "Success!"

    def raise_readtimeout_three_times():
        if arg_dict["retry_count"] == 3:
            return success_message
        else:
            arg_dict["retry_count"] += 1
            raise requests.exceptions.ReadTimeout()

    expected_wait_time = get_wait_time(None, retries=2)
    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        assert retry_for_status(raise_readtimeout_three_times) == success_message
        m.assert_called_with(expected_wait_time)


def test_retry_for_status_retry_works_custom_exception():
    arg_dict = {"retry_count": 0}
    success_message = "Success!"

    class CustomException(Exception):
        pass

    def raise_readtimeout_three_times():
        if arg_dict["retry_count"] == 1:
            return success_message
        else:
            arg_dict["retry_count"] += 1
            raise CustomException()

    expected_wait_time = get_wait_time(None, retries=0)
    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        assert (
            retry_for_status(raise_readtimeout_three_times, exceptions_to_retry=(CustomException,))
            == success_message
        )
        m.assert_called_with(expected_wait_time)


def test_retry_for_status_with_jira_non_caught_error():
    def raise_401():
        raise JIRAError(text="Generic Error Testing", status_code=401)

    def raise_429():
        raise JIRAError(text="Generic Error Testing", status_code=429)

    with pytest.raises(JIRAError):
        retry_for_status(raise_401)

    with pytest.raises(JIRAError):
        retry_for_status(raise_429(), status_codes_to_retry=[])


def test_batch_iterable():
    total_data_size = 85926
    batch_size = 62
    data = [i for i in range(total_data_size)]

    recompiled_batches = []
    for batch in batch_iterable(data, batch_size=batch_size):
        # Assert we are getting proper batch sizes
        assert len(batch) <= batch_size
        recompiled_batches.extend(batch)

    # Assert we got ALL the data
    assert len(recompiled_batches) == total_data_size

    # Assert that data is returned matching and in order
    for i in range(total_data_size):
        assert data[i] == recompiled_batches[i]


def test_batch_iterable_by_bytes_size():
    number_of_chars = 41234
    data = ['a' for i in range(number_of_chars)]
    total_data_size_in_bytes = sum(sys.getsizeof(i) for i in data)

    # NOTE: This values should be greater than 50
    batch_bytes_size = 500

    batch_numbers = 0
    recompiled_batches = []
    for batch in batch_iterable_by_bytes_size(data, batch_byte_size=batch_bytes_size):
        # Assert we are getting proper batch sizes
        assert sys.getsizeof(batch) > 0
        assert sys.getsizeof(batch) <= batch_bytes_size
        batch_numbers += 1
        recompiled_batches.extend(batch)

    # Assert we got ALL the data
    assert len(recompiled_batches) == len(data)
    assert batch_numbers == math.ceil(total_data_size_in_bytes / batch_bytes_size)


def test_get_object_bytes_size():
    test_object = {'a': 1, 'b': 2, 'c': 3}

    object_size = get_object_bytes_size(test_object)

    manually_derived_size = 0
    for key in test_object.keys():
        manually_derived_size += sys.getsizeof(key)
        manually_derived_size += sys.getsizeof(test_object[key])
    assert manually_derived_size == object_size


def test_get_object_bytes_size_nested_object():
    test_object = {'a': 1, 'b': 2, 'c': {'d': 4}}

    object_size = get_object_bytes_size(test_object)

    manually_derived_size = 0
    for key in test_object.keys():
        manually_derived_size += sys.getsizeof(key)
        value = test_object[key]
        if type(value) == dict:
            for _key in value.keys():
                manually_derived_size += sys.getsizeof(_key)
                _value = value[_key]
                manually_derived_size += sys.getsizeof(_value)
        else:
            manually_derived_size += sys.getsizeof(value)

    assert manually_derived_size == object_size


def test_format_date_to_jql():
    min_datetime = datetime.min
    assert format_date_to_jql(min_datetime) == '"0001-01-01"'

    _american_revolution_datetime = datetime(1776, 7, 4)
    assert format_date_to_jql(_american_revolution_datetime) == '"1776-07-04"'

    _datetime_pre_unix_epoch_time = datetime(1969, 12, 31)
    assert format_date_to_jql(_datetime_pre_unix_epoch_time) == '"1969-12-31"'

    _datetime_after_unix_epoch_time = datetime(2003, 3, 12)
    assert format_date_to_jql(_datetime_after_unix_epoch_time) == '"2003-03-12"'

    _datetime_after_unix_epoch_time = datetime(2024, 12, 18)
    assert format_date_to_jql(_datetime_after_unix_epoch_time) == '"2024-12-18"'

    _future_datetime = datetime(3333, 1, 4)
    assert format_date_to_jql(_future_datetime) == '"3333-01-04"'


def test_issue_metadata_serializer_single_object():
    # Create an IssueMetadata Object with values we can
    # verify after we do a "round trip" of serializing
    # and deserializing
    id = "1"
    key = "PROJ-1"
    _datetime = datetime(2023, 12, 29)
    epic_link_field_issue_key = "EPIC_LINK_FIELD_ISSUE_KEY"
    parent_field_issue_key = "PARENT_FIELD_ISSUE_KEY"
    issue_metadata = IssueMetadata(
        id=id,
        key=key,
        updated=_datetime,
        epic_link_field_issue_key=epic_link_field_issue_key,
        parent_field_issue_key=parent_field_issue_key,
    )

    # Serialize the object to a string
    serialized_issue_metadata = IssueMetadata.to_json_str(issue_metadata)
    assert type(serialized_issue_metadata) == str

    # Deserialize it back to memory, verify that it is the proper type
    deserialized_issue_metadata = IssueMetadata.from_json(serialized_issue_metadata)
    assert type(deserialized_issue_metadata) == IssueMetadata

    # Assert that all the data is the same after a serialization/deserialization round trip
    assert deserialized_issue_metadata.id == id
    assert deserialized_issue_metadata.key == key
    assert deserialized_issue_metadata.updated == _datetime
    assert deserialized_issue_metadata.epic_link_field_issue_key == epic_link_field_issue_key
    assert deserialized_issue_metadata.parent_field_issue_key == parent_field_issue_key


def test_issue_metadata_serializer_list_of_objects():
    # Generate a list of IssueMetadata to serialize/deserialize
    issue_metadata_list: list[IssueMetadata] = []
    lower_bound_day = 1
    upper_bound_day = 30
    for day in range(lower_bound_day, upper_bound_day):
        id = f"{day}"
        key = f"PROJ-{day}"
        _datetime = datetime(2023, 12, day)
        epic_link_field_issue_key = f"EPIC_LINK_FIELD_ISSUE_KEY_{day}"
        parent_field_issue_key = f"PARENT_FIELD_ISSUE_KEY_{day}"

        issue_metadata_list.append(
            IssueMetadata(
                id=id,
                key=key,
                updated=_datetime,
                epic_link_field_issue_key=epic_link_field_issue_key,
                parent_field_issue_key=parent_field_issue_key,
            )
        )

    # Verify we can serialize it as a list
    serialized_issue_metadata_list = IssueMetadata.to_json_str(issue_metadata_list)
    assert type(serialized_issue_metadata_list) == str

    # Verify that we can deserialize it as a list of the same size and of the same type
    deserialized_issue_metadata_list = IssueMetadata.from_json(serialized_issue_metadata_list)
    assert type(deserialized_issue_metadata_list) == list
    assert len(deserialized_issue_metadata_list) == len(issue_metadata_list)
    for item in deserialized_issue_metadata_list:
        assert type(item) == IssueMetadata

    for i in range(len(issue_metadata_list)):
        issue_metadata_pre = issue_metadata_list[i]
        issue_metadata_post = deserialized_issue_metadata_list[i]

        assert issue_metadata_pre.id == issue_metadata_post.id
        assert issue_metadata_pre.key == issue_metadata_post.key
        assert issue_metadata_pre.updated == issue_metadata_post.updated
        assert (
            issue_metadata_pre.epic_link_field_issue_key
            == issue_metadata_post.epic_link_field_issue_key
        )
        assert (
            issue_metadata_pre.parent_field_issue_key == issue_metadata_post.parent_field_issue_key
        )


def test_hash_files():
    filepath = 'a/path/to/a/file.py'
    hashed_filepath = hash_filename(filepath)
    assert hashed_filepath == '0cc175b9c0f1/d6fe1d0be634/01b6e20344b6/0cc175b9c0f1/02b202aa0962.py'

    # Test when file has additional . in it
    filepath = 'a/path/to.a/file.py'
    hashed_filepath = hash_filename(filepath)
    assert hashed_filepath == '0cc175b9c0f1/d6fe1d0be634/18449aad14e9/02b202aa0962.py'

    # Test when file has a bunch of additional . in it
    filepath = 'a.path/to.a/file.py'
    hashed_filepath = hash_filename(filepath)
    assert hashed_filepath == '8bd1e22aeb9f/18449aad14e9/02b202aa0962.py'



from datetime import datetime
from typing import Dict, Generator, List, Optional

import requests

from jf_ingest import logging_helper
from jf_ingest.config import AzureDevopsAuthConfig, GithubAuthConfig
from jf_ingest.utils import retry_for_status, retry_session

ADO_DEFAULT_API_URL = 'https://dev.azure.com'


class AzureDevopsClient:
    '''
    Similar to GithubClient and BitbucketCloudClient in that this a requests-based
    client (i.e., we're not using a specialized library).

    Makes requests to the Azure DevOps API, handling that API's specific handling
    of authentication, pagination, retries, rate limiting, etc.
    '''

    def __init__(
        self,
        github_auth_config: AzureDevopsAuthConfig,
        **kwargs,
    ):
        """Wrapper class for the DevOps API endpoint.

        Args:
            github_auth_config (GithubAuthConfig): A fully fleshed out GithubAuthConfig, with the api_version set
        """
        self.base_url = github_auth_config.base_url or ADO_DEFAULT_API_URL
        self.token = github_auth_config.token
        self.session = github_auth_config.session or retry_session(**kwargs)
        self.api_version = github_auth_config.api_version
        if self.api_version:
            self.api_version = github_auth_config.api_version
            self.common_request_params = {
                'api-version': self.api_version,
            }

        # Basic authentication, using no username and the PAT as the password
        self.session.auth = ('', self.token)
        self.session.verify = True
        self.session.headers.update({'Accept': 'application/json'})

    def _get_raw_result(
        self, url: str, params: Optional[Dict] = None, ignore404: Optional[bool] = False
    ) -> tuple[requests.Response, str]:
        """
        Returns a tuple: (the_raw_result, continuation_token_if_applicable)
        If there are more results to be fetched, some but not all API methods will return a continuation_token.

        This function WILL raise errors (like 429s) and should be called with the retry_for_status wrapper

        Args:
            url (str): A URL to hit with a GET request.
            params (Optional[Dict], optional): Params to hit the GET call with. Defaults to None.
            ignore404 (Optional[bool], optional): If provided, 404 errors will be ignored. Defaults to False.

        Returns:
            tuple[requests.Response, str]: Returns the response and a str that represents the continuation token
        """
        if params is None:
            params = {}
        try:
            result = self.session.get(url=url, params=dict(self.common_request_params, **params))
            result.raise_for_status()
            continuation_token = result.headers.get('x-ms-continuationtoken')
            return result, continuation_token
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404 and ignore404:
                logging_helper.send_to_agent_log_file(f'Caught a 404 for {url} - ignoring')
                return None, None
            raise

    def get_json(
        self, url: str, params: Optional[Dict] = None, ignore404: Optional[bool] = False
    ) -> tuple[Dict, str]:
        """
        Returns a tuple: (the_json_result, continuation_token_if_applicable)
        If there are more results to be fetched, some but not all API methods will return a continuation_token

        Args:
            url (str): URL to query
            params (Optional[Dict], optional): Params to query URL with. Defaults to None.
            ignore404 (Optional[bool], optional): Boolean to mark if we should ignore any raised 404s. Defaults to False.

        Returns:
            tuple[Dict, str]: A tuple of the raw Dict result and the continuation token
        """
        if params is None:
            params = {}

        # Make request via the _get_raw_result function, wrapped in the retry_for_status util function
        # Save results initially as a tuple so we can unpack it later, giving us cleaner type hinting
        raw_result_tuple: tuple[requests.Response, str] = retry_for_status(
            self._get_raw_result, url=url, params=params, ignore404=ignore404
        )
        result, continuation_token = raw_result_tuple
        if result:
            return result.json(), continuation_token
        else:
            return {}, None

    def get_single_object(
        self, url: str, params: Optional[Dict] = None, ignore404: Optional[bool] = False
    ) -> Dict:
        """Get a single object from the ADO API

        Args:
            url (str): URL to query
            params (Optional[Dict], optional): Params to query URL with. Defaults to None.
            ignore404 (Optional[bool], optional): Boolean to mark if we should ignore any raised 404s. Defaults to False.

        Returns:
            Dict: a single JSON result returned from the API
        """
        if params is None:
            params = {}
        result, _ = self.get_json(url, params, ignore404)
        return result

    def get_single_page(
        self,
        url: str,
        params: Optional[Dict] = None,
        ignore404: Optional[bool] = False,
        result_container: Optional[str] = 'value',
    ) -> List[Dict]:
        """
        For API methods that do NOT implement pagination -- all results come back
        in a single page

        Args:
            url (str): URL to query
            params (Optional[Dict], optional): Params to query URL with. Defaults to None.
            ignore404 (Optional[bool], optional): Boolean to mark if we should ignore any raised 404s. Defaults to False.
            result_container (Optional[str], optional): A string indicating the key within the returned dictionary that holds the list of results. Defaults to 'value'.

        Returns:
            List[Dict]: A list of dictionary objects returned from the API
        """
        if params is None:
            params = {}
        page, _ = self.get_json(url, params, ignore404)
        if page and result_container in page:
            return page[result_container]
        else:
            return []

    def get_all_pages_using_skip_and_top(
        self,
        url: str,
        params: Optional[Dict] = None,
        ignore404: Optional[bool] = False,
        result_container: Optional[str] = 'value',
        skip: Optional[int] = 0,
        top: Optional[int] = 100,
    ) -> Generator[Dict, None, None]:
        """For API methods that implement resultset pagination using $skip and $top. To paginate,
        use $top=100 and $skip=0, then $skip=100, then $skip=200, etc.

        Args:
            url (str): URL to query
            params (Optional[Dict], optional): Params to query URL with. Defaults to None.
            ignore404 (Optional[bool], optional): Boolean to mark if we should ignore any raised 404s. Defaults to False.
            result_container (Optional[str], optional): A string indicating the key within the returned dictionary that holds the list of results. Defaults to 'value'.
            skip (Optional[int], optional): tells the API to skip this many matching entities. Defaults to 0.
            top (Optional[int], optional): tells the API to return at most this many matching entities. Defaults to 100.

        Yields:
            Generator[Dict, None, None]: Yields a dictionary for each result
        """
        if params is None:
            params = {}
        while True:
            params.update({'$skip': skip, '$top': top})
            page, continuation_token = self.get_json(url, params, ignore404)
            assert not continuation_token, (
                "Inside of get_all_pages_using_skip_and_top, but a continuation_token was included in the response; "
                "do you mean to use get_all_pages_using_pagination_token instead?"
            )
            if not page or result_container not in page:
                return
            for value in page[result_container]:
                yield value
            if 'count' in page and page['count'] == top:
                skip += top
            # NOTE: Fot the Diffs endpoint, the "count" is obscured as this
            # changeCounts values and we have to handle it differently
            elif 'changeCounts' in page:
                # Get TOTAL count of changed files (this can be bigger than page size (top))
                count_from_change_counts = sum(page['changeCounts'].values())
                # skip ahead a page
                skip += top
                # If we've skipped a page ahead and we're still below
                # the total, return
                #
                # If we've skipped a page ahead and we're equal to or greater than
                # the total files, return
                if count_from_change_counts > skip:
                    continue
                else:
                    return
            else:
                return

    def get_all_pages_using_pagination_token(
        self,
        url: str,
        params: Optional[Dict] = None,
        ignore404: Optional[bool] = False,
        result_container: Optional[str] = 'value',
    ) -> Generator[Dict, None, None]:
        """For API methods that implement resultset pagination using a continuation token
        To paginate, we issue subsequent requests using the continuation_token returned

        Args:
            url (str): URL to query
            params (Optional[Dict], optional): Params to query URL with. Defaults to None.
            ignore404 (Optional[bool], optional): Boolean to mark if we should ignore any raised 404s. Defaults to False.
            result_container (Optional[str], optional): A string indicating the key within the returned dictionary that ho

        Yields:
            Generator[Dict, None, None]: Yields a dictionary for each result
        """
        if params is None:
            params = {}
        while True:
            page, continuation_token = self.get_json(url, params, ignore404)
            if not page or result_container not in page:
                return
            for value in page[result_container]:
                yield value
            if continuation_token:
                params.update({'continuationToken': continuation_token})
            else:
                return

    def get_all_repos(self, org_name: str) -> List[Dict]:
        """Get all repositories from the ADO API
        Somehow, this API endpoint doesn't support paging? The only option is to get all the repositories all at once
        https://learn.microsoft.com/en-us/rest/api/azure/devops/git/repositories/list?view=azure-devops-rest-7.1&tabs=HTTP

        Args:
            org_name (str): The organization to query for

        Returns:
            List[Dict]: A list of repo dictionaries from ADO
        """
        url = f'{self.base_url}/{org_name}/_apis/git/repositories?api-version={self.api_version}'
        return self.get_single_page(url)

    def get_graph_users(self, org_name: str) -> Generator[Dict, None, None]:
        """A Generator for getting user data from the ADO API

        Args:
            org_name (str): The org to query for

        Yields:
            Generator[Dict, None, None]: A raw User dictionary object
        """
        url = f"{self.base_url}/{org_name}/_apis/graph/users"
        return self.get_all_pages_using_pagination_token(
            # XXX: This API is not yet GA? Hopefully down the road we'll be able to do away
            # with the URL and api-version overrides
            url.replace('dev.azure.com', 'vssps.dev.azure.com'),
            params={'api-version': '7.0-preview.1'} if self.api_version == '7.0' else None,
        )

    def get_teams(self, org_name: str) -> Generator[Dict, None, None]:
        """A Generator for getting team data from the ADO API

        Args:
            org_name (str): The org to query for

        Yields:
            Generator[Dict, None, None]: A raw Team dictionary object
        """
        url = f"{self.base_url}/{org_name}/_apis/graph/groups"
        return self.get_all_pages_using_pagination_token(
            # XXX: This API is not yet GA? Hopefully down the road we'll be able to do away
            # with the URL and api-version overrides
            url.replace('dev.azure.com', 'vssps.dev.azure.com'),
            params={'api-version': '7.0-preview.1'} if self.api_version == '7.0' else None,
        )

    def get_team_users(self, org_name: str, team_descriptor: str) -> List[Dict]:
        """Get a list of users for a given team

        Args:
            org_name (str): An organization name
            team_id (str): A team ID

        Returns:
            List[Dict]: A list of members for a team
        """
        url = f"{self.base_url}/{org_name}/_apis/Graph/Memberships/{team_descriptor}"
        params = {'direction': 'down'}
        if self.api_version == '7.0':
            params['api-version'] = '7.0-preview.1'
        team_memberships = self.get_single_page(
            # XXX: This API is not yet GA? Hopefully down the road we'll be able to do away
            # with the URL and api-version overrides
            url.replace('dev.azure.com', 'vssps.dev.azure.com'),
            params=params,
        )
        return [
            self.get_json(
                # XXX: This API is not yet GA? Hopefully down the road we'll be able to do away
                # with the api-version override
                team_membership['_links']['member']['href'],
                params={'api-version': '7.0-preview.1'} if self.api_version == '7.0' else None,
            )[0]
            for team_membership in team_memberships
        ]

    def get_commits(
        self,
        org_name: str,
        project_name: str,
        repo_id: str,
        branch_name: str,
        from_date: Optional[datetime] = None,
        to_date: Optional[datetime] = None,
    ) -> Generator[Dict, None, None]:
        """Get commits for a branch (within a repo, within a project, and within an Org), with the ability to filter in between dates

        Args:
            org_name (str): An organization name
            project_name (str): A Project name. A project name is extracted from the repo URL. See _project_name_from_repo
            repo_id (str): A Repo ID
            branch_name (str): A Branch Name
            from_date (Optional[datetime], optional): search will only return items created AFTER this date. Defaults to None.
            to_date (Optional[datetime], optional): search will only return items created BEFORE this date. Defaults to None.

        Yields:
            Generator[Dict, None, None]: A raw Commit Dictionary
        """
        url = f"{self.base_url}/{org_name}/{project_name}/_apis/git/repositories/{repo_id}/commits?api-version={self.api_version}"
        params = {
            'searchCriteria.itemVersion.version': branch_name,
            'searchCriteria.itemVersion.versionType': 'branch',
        }
        if from_date:
            params.update({'searchCriteria.fromDate': from_date.strftime('%m/%d/%Y %H:%M:%S %p')})
        if to_date:
            params.update({'searchCriteria.toDate': to_date.strftime('%m/%d/%Y %H:%M:%S %p')})

        return self.get_all_pages_using_skip_and_top(url, params)

    def get_commit(self, org_name: str, project_name: str, repo_id: str, commit_id: str) -> Dict:
        """Get a single commit

        Args:
            org_name (str): An organization name
            project_name (str): A Project name. A project name is extracted from the repo URL. See _project_name_from_repo
            repo_id (str): A Repo ID
            branch_name (str): A Branch Name
            commit_id (str): a commit ID to search for

        Returns:
            Dict: A raw Commit Dictionary
        """
        url = f"{self.base_url}/{org_name}/{project_name}/_apis/git/repositories/{repo_id}/commits/{commit_id}?api-version={self.api_version}"
        return self.get_single_object(url)

    def get_branches(
        self, org_name: str, project_name: str, repo_id: str, filter_startswith: str = None
    ) -> Generator[Dict, None, None]:
        """Get a generator of branches

        Args:
            org_name (str): An org name
            project_name (str): A Project name. A project name is extracted from the repo URL. See _project_name_from_repo
            repo_id (str): A repository ID
            filter_startswith (str, optional): A string to filter branches with. Defaults to None.

        Yields:
            Generator[Dict, None, None]: A raw branch in the form of a dictionary
        """
        url = f"{self.base_url}/{org_name}/{project_name}/_apis/git/repositories/{repo_id}/refs?api-version={self.api_version}"
        if filter_startswith:
            return self.get_all_pages_using_pagination_token(
                url, params={'filter': f'heads/{filter_startswith}'}
            )
        else:
            return self.get_all_pages_using_pagination_token(url)

    def get_pull_request_iterations(
        self, org_name: str, project_name: str, repo_id: str, pr_id: int
    ) -> List[Dict]:
        url = f'{self.base_url}/{org_name}/{project_name}/_apis/git/repositories/{repo_id}/pullRequests/{pr_id}/iterations?api-version={self.api_version}'
        return self.get_single_page(url)

    def get_pull_request_labels(
        self, org_name: str, project_name: str, repo_id: str, pr_id: int
    ) -> List[Dict]:
        url = f"{self.base_url}/{org_name}/{project_name}/_apis/git/repositories/{repo_id}/pullRequests/{pr_id}/labels?api-version={self.api_version}"
        return self.get_single_page(url)

    def get_pull_request_changes_counts(
        self, org_name: str, repo_id: str, base_sha: str, target_sha: str
    ) -> Dict:
        url = f"{self.base_url}/{org_name}/_apis/git/repositories/{repo_id}/diffs/commits?api-version={self.api_version}"
        params = {
            'baseVersion': base_sha,
            'baseVersionType': 'commit',
            'targetVersion': target_sha,
            'targetVersionType': 'commit',
        }
        return self.get_json(url, params)[0]['changeCounts']

    def get_pull_request_diff(
        self, org_name: str, repo_id: str, base_sha: str, target_sha: str
    ) -> List[Dict]:
        url = f"{self.base_url}/{org_name}/_apis/git/repositories/{repo_id}/diffs/commits?api-version={self.api_version}"
        params = {
            'baseVersion': base_sha,
            'baseVersionType': 'commit',
            'targetVersion': target_sha,
            'targetVersionType': 'commit',
        }
        return self.get_all_pages_using_skip_and_top(url, params, result_container='changes')

    def get_pull_requests(
        self,
        org_name: str,
        project_name: str,
        repo_id: str,
        start_window: datetime,
        end_window: datetime,
        status: str = 'active',
        filter_by: str = 'created',
    ) -> Generator[Dict, None, None]:
        """Generate pull request dictionaries, pulled from the ADO API

        Args:
            org_name (str): An organization name
            project_name (str): A Project name. A project name is extracted from the repo URL. See _project_name_from_repo
            repo_id (str): A repository ID
            start_cursor (int): The cursor of where in pagination we should start. Defaults to 0.

        Yields:
            Generator[Dict, None, None]: A raw pull request dictionary
        """

        url = f"{self.base_url}/{org_name}/{project_name}/_apis/git/repositories/{repo_id}/pullrequests?api-version={self.api_version}&searchCriteria.status={status}"
        if start_window or end_window:
            url += f'&searchCriteria.queryTimeRangeType={filter_by}'
            if start_window:
                url += f'&searchCriteria.minTime={start_window}'
            if end_window:
                url += f'&searchCriteria.maxTime={end_window}'
        return self.get_all_pages_using_skip_and_top(url)

    def get_pr_comment_threads(
        self, org_name: str, project_name: str, repo_id: str, pr_id: int
    ) -> List[Dict]:
        """Gets a list of PR Comments for a Pull Request

        Args:
            org_name (str): An organization name
            project_name (str): A Project name. A project name is extracted from the repo URL. See _project_name_from_repo
            repo_id (str): A repository ID
            pr_id (int): A pull request ID (pull request Number)

        Returns:
            List[Dict]: A list of comment objects for a PR
        """
        url = f"{self.base_url}/{org_name}/{project_name}/_apis/git/repositories/{repo_id}/pullRequests/{pr_id}/threads?api-version={self.api_version}"
        return self.get_single_page(url)

    def get_pr_commits(
        self, org_name: str, project_name: str, repo_id: str, pr_id: int
    ) -> Generator[Dict, None, None]:
        """Get a list of all commits associated with a Pull Request

        Args:
            org_name (str): An organization name
            project_name (str): A Project name. A project name is extracted from the repo URL. See _project_name_from_repo
            repo_id (str): A repository ID
            pr_id (int): A pull request ID (pull request Number)

        Yields:
            Generator[Dict, None, None]: A commit object associated with the PR
        """
        url = f"{self.base_url}/{org_name}/{project_name}/_apis/git/repositories/{repo_id}/pullRequests/{pr_id}/commits?api-version={self.api_version}"
        return self.get_all_pages_using_pagination_token(url)

    def get_pr_labels(
        self, org_name: str, project_name: str, repo_id: str, pr_id: int
    ) -> List[Dict]:
        """Get a list of PR labels for a specific PR

        Args:
            org_name (str): An organization name
            project_name (str): A Project name. A project name is extracted from the repo URL. See _project_name_from_repo
            repo_id (str): A repository ID
            pr_id (int): A pull request ID (pull request Number)

        Returns:
            List[Dict]: A list of label objects for a PR
        """
        url = f"{self.base_url}/{org_name}/{project_name}/_apis/git/repositories/{repo_id}/pullRequests/{pr_id}/labels?api-version={self.api_version}"
        return self.get_single_page(url)

    def get_diffs_between_shas(
        self, org_name: str, repo_id: str, base_sha: str, target_sha: str
    ) -> List[Dict]:
        """Derive the additions, deletions, and changed_files between two shas (commits). It does so by
        using the "diffs" API endpoint:

        https://learn.microsoft.com/en-us/rest/api/azure/devops/git/diffs/get?view=azure-devops-rest-7.1&tabs=HTTP

        THIS FUNCTION IS USED TO RECONSTRUCT ADDITIONS/DELETIONS/CHANGED FOR A PULL REQUEST!

        Unfortunately the ADO API doesn't provide any way to get the LOC additions/deletions -- they're
        not returned as attributes of a PR, or of a commit, nor do they come back in the response from
        the "diffs" endpoint.

        So all we can get is the number of changed files. We get this by calling the "diffs" endpoint,
        comparing the tip of the PR's branch with the base branch.

        Args:
            org_name (str): An organization name
            repo_id (str): A repository that contains the base and target_shas
            base_sha (str): A base sha to compare with
            target_sha (str): A target sha to compare to

        Returns:
            List[Dict]: A list of differences, by file and action. See link to API documenation (in description) for more
        """
        url = f"{self.base_url}/{org_name}/_apis/git/repositories/{repo_id}/diffs/commits?api-version={self.api_version}"
        params = {
            'baseVersion': base_sha,
            'baseVersionType': 'commit',
            'targetVersion': target_sha,
            'targetVersionType': 'commit',
        }
        return self.get_single_page(url, params, result_container='changes')



import enum
from dataclasses import dataclass


class EventState(str, enum.Enum):
    START = 'START'
    SUCCESS = 'SUCCESS'
    ERROR = 'ERROR'


class IngestType(str, enum.Enum):
    JIRA = 'JIRA_INGEST'
    GIT = 'GIT_INGEST'


class JiraIngestEventNames:
    GET_JIRA_DATA = 'get_jira_data'
    GET_JIRA_CONNECTION = 'get_jira_connection'
    GET_JIRA_PROJECTS = 'get_jira_projects'
    GET_JIRA_FIELDS = 'get_jira_fields'
    GET_JIRA_USERS = 'get_jira_users'
    GET_JIRA_RESOLUTIONS = 'get_jira_resolutions'
    GET_JIRA_ISSUE_TYPES = 'get_jira_issue_types'
    GET_JIRA_LINK_TYPES = 'get_jira_link_types'
    GET_JIRA_PRIORITIES = 'get_jira_priorities'
    GET_JIRA_STATUSES = 'get_jira_statuses'
    GET_JIRA_BOARD_AND_SPRINTS = 'get_jira_board_and_sprints'
    GET_JIRA_ISSUES = 'get_jira_issues'
    GET_JIRA_WORKLOGS = 'get_jira_worklogs'


class GitIngestionEventNames:
    GET_GIT_DATA = 'get_git_data'
    GET_GIT_ORGANIZATIONS = 'get_git_organizations'
    GET_GIT_USERS = 'get_git_users'
    GET_GIT_TEAMS = 'get_git_teams'
    GET_GIT_REPOS = 'get_git_repos'
    GET_GIT_BRANCHES = 'get_git_branches'
    GET_GIT_COMMITS = 'get_git_commits'
    GET_GIT_PULL_REQUESTS = 'get_git_pull_requests'


@dataclass
class IngestEvent:
    company_slug: str
    # This will be filled out by child classes
    ingest_type: IngestType
    # this should reflect what we're doing within the ingest component
    # "get_git_repos", "get_jira_issues", etc.
    event_name: str

    event_state: EventState = None
    # Only applicable if is_error is true
    error_message: str = ''

    @staticmethod
    def from_dict(event: dict) -> "IngestEvent":
        if ingest_type := event.get(f'ingest_type'):
            if ingest_type == IngestType.JIRA.value:
                return JiraIngestEvent(**event)
            elif ingest_type == IngestType.GIT.value:
                return GitIngestEvent(**event)


@dataclass
class JiraIngestEvent(IngestEvent):
    ingest_type = IngestType.JIRA


@dataclass
class GitIngestEvent(IngestEvent):
    ingest_type = IngestType.GIT
    git_instance: str = ''
    git_provider: str = ''



import logging
from datetime import datetime
from enum import Enum
from typing import Generator

from jira import JIRA

from jf_ingest import diagnostics, logging_helper
from jf_ingest.config import (
    IngestionConfig,
    IngestionType,
    IssueDownloadingResult,
    IssueMetadata,
    JiraAuthMethod,
    JiraDownloadConfig,
)
from jf_ingest.constants import Constants
from jf_ingest.events.models import IngestType, JiraIngestEvent, JiraIngestEventNames
from jf_ingest.events.utils import emit_event_to_jellyfish_context_manager
from jf_ingest.file_operations import IngestIOHelper, SubDirectory
from jf_ingest.jf_jira.auth import get_jira_connection
from jf_ingest.jf_jira.downloaders import (
    download_all_issue_metadata,
    download_boards_and_sprints,
    download_fields,
    download_issuelinktypes,
    download_issues,
    download_issues_from_new_sync,
    download_issuetypes,
    download_priorities,
    download_projects_and_versions_and_components,
    download_resolutions,
    download_statuses,
    download_users,
    download_worklogs,
    get_ids_from_difference_of_issue_metadata,
    get_issue_ids_for_rekeys_and_deletes,
    get_jira_search_batch_size,
    pull_all_jira_issues_by_date,
    pull_jira_issues_by_jira_ids,
)
from jf_ingest.telemetry import add_telemetry_fields, jelly_trace, record_span
from jf_ingest.utils import (
    batch_iterable_by_bytes_size,
    get_jellyfish_company_slug,
    init_jf_ingest_run,
)

logger = logging.getLogger(__name__)


def _generate_jira_event(event_name: str) -> JiraIngestEvent:
    return JiraIngestEvent(
        company_slug=get_jellyfish_company_slug(),
        event_name=event_name,
        ingest_type=IngestType.JIRA,
    )


class JiraObject(Enum):
    JiraFields = "jira_fields"
    JiraProjectsAndVersions = "jira_projects_and_versions"
    JiraUsers = "jira_users"
    JiraResolutions = "jira_resolutions"
    JiraIssueTypes = "jira_issuetypes"
    JiraLinkTypes = "jira_linktypes"
    JiraPriorities = "jira_priorities"
    JiraBoards = "jira_boards"
    JiraSprints = "jira_sprints"
    JiraBoardSprintLinks = "jira_board_sprint_links"
    JiraIssues = "jira_issues"
    JiraIssuesIdsDownloaded = "jira_issue_ids_downloaded"
    JiraIssuesIdsDeleted = "jira_issue_ids_deleted"
    JiraWorklogs = "jira_worklogs"
    JiraStatuses = "jira_statuses"


@jelly_trace()
def load_issues_in_batches(
    issues_to_download: Generator[dict, None, None],
    ingest_io_helper: IngestIOHelper,
    ingest_config: IngestionConfig,
    batch_number_start: int = 0,
) -> IssueDownloadingResult:
    """given a generator object for issues to download; save them to disk in batches according to maximum size

    Args:
        issues_to_download: Generator object for issues to download
        ingest_io_helper: IngestIOHelper object for doing file operations
        ingest_config (IngestionConfig): A dataclass that holds several different configuration args for this task
        batch_number_start: starting index for the batch number (jira_issuesN.json). Used if you call this function multiple times

    Return:
        IssueDownloadingResult: A dataclass that holds the results of the download;
            - the ids for jira issues retrieved,
            - jira_ids for parents of those retrieved issues,
            - and the total number of batches it took
    """
    total_issue_batches = 0
    all_downloaded_issue_ids = set()
    collected_parent_ids = set()

    for batch_number, batch_issues in enumerate(
        batch_iterable_by_bytes_size(
            issues_to_download, batch_byte_size=Constants.JIRA_ISSUES_UNCOMPRESSED_FILE_SIZE
        ),
        start=batch_number_start,
    ):
        logging_helper.send_to_agent_log_file(
            f'Saving {len(batch_issues)} issues as batch number {batch_number}',
        )
        all_downloaded_issue_ids.update(set(issue['id'] for issue in batch_issues))
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraIssues.value,
            subdirectory=SubDirectory.JIRA,
            json_data=batch_issues,
            batch_number=batch_number,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
            enable_telemetry=False,
        )
        total_issue_batches += 1

        # we need to download parents of issues we just fetched
        collected_parent_ids.update(
            set(
                issue['fields']['parent']['id']
                for issue in batch_issues
                if 'parent' in issue['fields'].keys()
            )
        )

    logger.info(
        f"Successfully saved {len(all_downloaded_issue_ids)} Jira Issues in "
        f"{total_issue_batches} separate batches, with each batch limited to "
        f"{round(Constants.JIRA_ISSUES_UNCOMPRESSED_FILE_SIZE / Constants.MB_SIZE_IN_BYTES)}MB per batch"
    )

    return IssueDownloadingResult(
        discovered_parent_ids=collected_parent_ids,
        downloaded_ids=all_downloaded_issue_ids,
        total_batches=total_issue_batches,
    )


@jelly_trace()
def pull_issues_old_metadata_path(
    jira_connect_or_fallback_connection: JIRA,
    jira_issues_batch_size: int,
    jira_config: JiraDownloadConfig,
    projects_and_versions: list[dict],
    ingest_io_helper: IngestIOHelper,
    ingest_config: IngestionConfig,
):
    """ "Default" old path for pulling issues
        get all metadata from jira, compare with metadata from JF, decide what to pull based on that
    Args:
        jira_connect_or_fallback_connection: either the basic connection or the special atlassian connect keys
        jira_issues_batch_size: batch size to use for actually downloading issues (much lower than the 10k for IDs)
        jira_config: jira config for jf_ingest from JF
        projects_and_versions: list of projects and versions to pull issues from
        ingest_io_helper: IngestIOHelper object for doing file operations
        ingest_config: full jf_ingest config
    """

    logging_helper.send_to_agent_log_file("Running 100 batched sync path")
    logging_helper.send_to_agent_log_file(
        f"All issue operations will use a batch_size of {jira_issues_batch_size}",
    )

    #######################################################################
    # Pull Issue Metadata
    #######################################################################
    issue_metadata_from_jira: list[IssueMetadata] = download_all_issue_metadata(
        jira_connection=jira_connect_or_fallback_connection,
        project_keys=[
            proj["key"]
            for proj in projects_and_versions
            if proj["key"] not in jira_config.exclude_projects
        ],
        pull_from=jira_config.pull_from,
        num_parallel_threads=jira_config.issue_download_concurrent_threads,
        batch_size=jira_issues_batch_size,
        recursively_download_parents=jira_config.recursively_download_parents,
    )

    issues_generator = download_issues(
        jira_connection=jira_connect_or_fallback_connection,
        full_redownload=jira_config.full_redownload,
        issue_download_concurrent_threads=jira_config.issue_download_concurrent_threads,
        jira_issues_batch_size=jira_issues_batch_size,
        issue_metadata_from_jellyfish=jira_config.jellyfish_issue_metadata,
        issue_metadata_from_jira=issue_metadata_from_jira,
        include_fields=jira_config.include_fields,
        exclude_fields=jira_config.exclude_fields,
    )

    all_downloaded_issue_ids = []
    total_issue_batches = 0

    for batch_number, batch_issues in enumerate(
        batch_iterable_by_bytes_size(
            issues_generator, batch_byte_size=Constants.JIRA_ISSUES_UNCOMPRESSED_FILE_SIZE
        )
    ):
        logging_helper.send_to_agent_log_file(
            f'Saving {len(batch_issues)} issues as batch number {batch_number}',
        )
        all_downloaded_issue_ids.extend([issue['id'] for issue in batch_issues])
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraIssues.value,
            subdirectory=SubDirectory.JIRA,
            json_data=batch_issues,
            batch_number=batch_number,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
            enable_telemetry=False,
        )
        total_issue_batches += 1

    add_telemetry_fields({'jira_downloaded_issue_count': len(all_downloaded_issue_ids)})
    add_telemetry_fields({'jira_downloaded_issue_batch_count': total_issue_batches})
    logger.info(
        f"Successfully saved {len(all_downloaded_issue_ids)} Jira Issues in {total_issue_batches} separate batches, "
        f"with each batch limited to "
        f"{round(Constants.JIRA_ISSUES_UNCOMPRESSED_FILE_SIZE / Constants.MB_SIZE_IN_BYTES)}MB per batch"
    )

    # Write Issues
    ingest_io_helper.write_json_to_local_or_s3(
        object_name=JiraObject.JiraIssuesIdsDownloaded.value,
        json_data=[int(issue_id) for issue_id in all_downloaded_issue_ids],
        subdirectory=SubDirectory.JIRA,
        save_locally=ingest_config.save_locally,
        upload_to_s3=ingest_config.upload_to_s3,
        enable_telemetry=False,
    )

    deleted_issue_ids = get_ids_from_difference_of_issue_metadata(
        jira_config.jellyfish_issue_metadata, issue_metadata_from_jira
    )

    return all_downloaded_issue_ids, deleted_issue_ids


@jelly_trace()
def pull_issues_new_sync_path(
    jira_connect_or_fallback_connection: JIRA,
    jira_config: JiraDownloadConfig,
    projects_and_versions: list[dict],
    jira_issues_batch_size: int,
    ingest_io_helper: IngestIOHelper,
    ingest_config: IngestionConfig,
):
    """Pulls issues using the "new" sync path, based on our ability to pull 10,000 jira issue IDs at one time
        as long as you only ask for `id` and `key` in the returned fields. Because we get no more fields we detect
        deletes as the absence of an issue on remote which we have local, and key changes as id match key mismatch
    Args:
        jira_connect_or_fallback_connection: either the basic connection or the special atlassian connect keys
        jira_config: jira config for jf_ingest from JF
        projects_and_versions: list of projects and versions to pull issues from
        jira_issues_batch_size: batch size to use for actually downloading issues (much lower than the 10k for IDs)
        ingest_io_helper: IngestIOHelper object for doing file operations
        ingest_config: full jf_ingest config

    """

    logging_helper.send_to_agent_log_file("Running 10k batched sync path")
    logger.info(
        f"Using {jira_issues_batch_size} as batch size for issue downloads. Full Redownload is {jira_config.full_redownload}"
    )

    ###########################################################################
    #
    # STEP ONE: Crawl across all projects to pull all Issue ID and Keys
    # to detect potential deletions, rekeys, or creations
    #
    ###########################################################################
    with record_span('pull_jira_issues_crawl'):
        issue_list_diff = get_issue_ids_for_rekeys_and_deletes(
            jira_connection=jira_connect_or_fallback_connection,
            jellyfish_issue_metadata=jira_config.jellyfish_issue_metadata,
            project_key_to_id={
                proj["key"]: proj["id"]
                for proj in projects_and_versions
                if proj["key"] not in jira_config.exclude_projects
            },
            pull_from=jira_config.pull_from,
        )
        ids_to_delete = issue_list_diff.ids_to_delete
        ids_to_download = issue_list_diff.ids_to_download

    ###########################################################################
    #
    # STEP TWO: Pull issues by their updated date
    #
    ###########################################################################
    project_id_to_pull_from: dict[str, datetime] = {
        str(proj["id"]): jira_config.project_id_to_pull_from.get(
            str(proj["id"]), jira_config.pull_from
        )
        if not jira_config.full_redownload
        else jira_config.pull_from
        for proj in projects_and_versions
        if proj["key"] not in jira_config.exclude_projects
    }
    logger.info(
        f'Attempting to pull all the full Jira Issues that have been updated since our last Jira Download across {len(project_id_to_pull_from)} projects.'
    )
    issues_by_updated_date_generator = pull_all_jira_issues_by_date(
        jira_connection=jira_connect_or_fallback_connection,
        project_key_to_pull_from=project_id_to_pull_from,
        num_parallel_threads=jira_config.issue_download_concurrent_threads,
        batch_size=jira_issues_batch_size,
        expand_fields=["renderedFields", "changelog"],
        include_fields=jira_config.include_fields,
        exclude_fields=jira_config.exclude_fields,
    )
    with record_span('pull_jira_issues_by_date'):
        issues_by_updated_date_results: IssueDownloadingResult = load_issues_in_batches(
            issues_by_updated_date_generator, ingest_io_helper, ingest_config
        )
        file_batch_cursor = issues_by_updated_date_results.total_batches
        logger.info(
            f'We have downloaded {len(issues_by_updated_date_results.downloaded_ids)} issues that have been updated or created since our last pull across {len(project_id_to_pull_from)} projects.'
        )
    ###########################################################################
    #
    # Step 3: Download the full Jira Issues for potentially rekeys and/or missing issues.
    # Also redownload issues that have been marked for redownload in our system
    #
    ###########################################################################
    # Extend this list with issues that need to be redownloaded for Jellyfish reasons
    ids_to_download.update(jira_config.jellyfish_issue_ids_for_redownload)
    # Next, exclude any IDs that we have just done a download for, via the updated date
    ids_to_download = ids_to_download.difference(issues_by_updated_date_results.downloaded_ids)
    logging_helper.send_to_agent_log_file(
        f'We need to download an additional {len(ids_to_download)} issues because they have been rekeyed or marked for redownload by Jellyfish'
    )

    # Finally, we have a "shopping list" of IDs to download
    issues_by_ids_generator = pull_jira_issues_by_jira_ids(
        jira_connection=jira_connect_or_fallback_connection,
        jira_ids=ids_to_download,
        num_parallel_threads=jira_config.issue_download_concurrent_threads,
        batch_size=jira_issues_batch_size,
        expand_fields=["renderedFields", "changelog"],
        include_fields=jira_config.include_fields,
        exclude_fields=jira_config.exclude_fields,
    )
    with record_span('pull_jira_issues_by_ids'):
        issues_by_ids_results: IssueDownloadingResult = load_issues_in_batches(
            issues_by_ids_generator,
            ingest_io_helper,
            ingest_config,
            batch_number_start=file_batch_cursor,
        )
        file_batch_cursor += issues_by_ids_results.total_batches

    # Now, combine all issue IDs that we've downloaded
    all_downloaded_ids: set[str] = set()
    all_downloaded_ids.update(issues_by_updated_date_results.downloaded_ids)
    all_downloaded_ids.update(issues_by_ids_results.downloaded_ids)

    # Then, combine the parents of the issues we downloaded by update date, and issues we downloaded by ID
    all_discovered_parent_ids: set[str] = set()
    all_discovered_parent_ids.update(issues_by_updated_date_results.discovered_parent_ids)
    all_discovered_parent_ids.update(issues_by_ids_results.discovered_parent_ids)

    # Now, find unique parents that we have NOT downloaded and that we need to
    existing_ids = set([jimd.id for jimd in jira_config.jellyfish_issue_metadata])
    all_known_ids = existing_ids.union(all_downloaded_ids)
    to_be_downloaded_parent_ids = all_discovered_parent_ids.difference(all_known_ids)

    ###########################################################################
    #
    # STEP FOUR: fetch parents
    # NOTE: For the general Jellyfish use case, we only need to go "one level deep" on parents,
    # i.e. pull the set of missing parents but NOT the parents of those parents (the "grandparents").
    # There is an optional feature, however, that will allow us to pull the parents of parents of parents
    # until we've pulled all parents. This is NOT the general use case, and will likely add a lot of computation
    # time to this function, so we advise you only use it in very specific use cases
    #
    ###########################################################################
    with record_span('pull_jira_issues_parents'):
        depth_level = 1
        while to_be_downloaded_parent_ids:
            logging_helper.send_to_agent_log_file(
                f'Attempting to pull more parents ({len(to_be_downloaded_parent_ids)} detected parents to pull). Some of them are {sorted(list(to_be_downloaded_parent_ids))[:30]} Depth level: {depth_level}'
            )
            issues_generator = download_issues_from_new_sync(
                jira_connection=jira_connect_or_fallback_connection,
                issue_download_concurrent_threads=jira_config.issue_download_concurrent_threads,
                jira_issues_batch_size=jira_issues_batch_size,
                issue_ids_to_download=to_be_downloaded_parent_ids,
                include_fields=jira_config.include_fields,
                exclude_fields=jira_config.exclude_fields,
            )
            with record_span('write_jira_parent_issues_batch_to_local_or_s3'):
                parents_batch_result: IssueDownloadingResult = load_issues_in_batches(
                    issues_to_download=issues_generator,
                    ingest_io_helper=ingest_io_helper,
                    ingest_config=ingest_config,
                    batch_number_start=file_batch_cursor,
                )

            all_known_ids.update(parents_batch_result.downloaded_ids)
            more_parent_ids = parents_batch_result.discovered_parent_ids - all_known_ids

            if jira_config.recursively_download_parents:
                # TODO: We could probably reduce the numbers of files we upload by grouping
                # all extra parents in joined files, but I don't think it's a big deal to upload
                # a large number of small files
                file_batch_cursor += parents_batch_result.total_batches
                to_be_downloaded_parent_ids = more_parent_ids
                depth_level += 1
            else:
                logging_helper.send_to_agent_log_file(
                    f'recursively_download_parents is set to {jira_config.recursively_download_parents}. Exiting execution'
                )
                break

    # Write Issues
    ingest_io_helper.write_json_to_local_or_s3(
        object_name=JiraObject.JiraIssuesIdsDownloaded.value,
        json_data=[int(issue_id) for issue_id in all_downloaded_ids],
        subdirectory=SubDirectory.JIRA,
        save_locally=ingest_config.save_locally,
        upload_to_s3=ingest_config.upload_to_s3,
        telemetry_span_name='write_jira_issue_ids_to_local_or_s3',
    )

    add_telemetry_fields({'jira_downloaded_issue_count': len(all_downloaded_ids)})

    return all_downloaded_ids, ids_to_delete


@jelly_trace
@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def load_and_push_jira_to_s3(ingest_config: IngestionConfig) -> bool:
    """Loads data from JIRA, Dumps it to disk, and then uploads that data to S3

    All configuration for this object is done via the JIRAIngestionConfig object

    Args:
        config (IngestionConfig): A dataclass that holds several different configuration args for this task

    Returns:
        bool: Returns True on Success
    """
    init_jf_ingest_run(ingestion_config=ingest_config)
    with emit_event_to_jellyfish_context_manager(
        _generate_jira_event(event_name=JiraIngestEventNames.GET_JIRA_DATA)
    ):
        company_slug = ingest_config.company_slug
        add_telemetry_fields({'company_slug': company_slug})
        return _run_load_and_push_to_s3(ingest_config=ingest_config)


def _run_load_and_push_to_s3(ingest_config: IngestionConfig) -> bool:
    """Inner function for load_and_push_jira_to_s3

    Args:
        config (IngestionConfig): A dataclass that holds several different configuration args for this task

    Returns:
        bool: Returns True on Success
    """
    logger.info('Beginning load_and_push_jira_to_s3')
    if not ingest_config.save_locally and not ingest_config.upload_to_s3:
        logger.error(
            f'Configuration error! Ingestion configuration must have either save_locally or upload_to_s3 set to True!'
        )
        raise Exception(
            'Save Locally and Upload to S3 are both set to False! Set one to true or no data will be saved!'
        )

    jira_config = ingest_config.jira_config
    if jira_config.skip_issues and jira_config.only_issues:
        raise Exception(
            'skip_issues and only_issues have both been set to True! This is an invalid JF Ingest object'
        )

    logging_helper.send_to_agent_log_file(f"Feature flags: {jira_config.feature_flags}")

    logger.info("Using local version of ingest")

    #######################################################################
    # SET UP JIRA CONNECTIONS (Basic and Potentially Atlassian Connect)
    #######################################################################
    with emit_event_to_jellyfish_context_manager(
        _generate_jira_event(event_name=JiraIngestEventNames.GET_JIRA_CONNECTION)
    ):
        jira_basic_connection = get_jira_connection(
            config=jira_config, auth_method=JiraAuthMethod.BasicAuth
        )

        jira_atlas_connect_connection = (
            get_jira_connection(config=jira_config, auth_method=JiraAuthMethod.AtlassianConnect)
            if JiraAuthMethod.AtlassianConnect in jira_config.available_auth_methods
            else None
        )
        # There is an ongoing effort to cut all things over to Atlassian Connect only,
        # but it is a piecewise migration for now.
        # OJ-29745
        jira_connect_or_fallback_connection = jira_basic_connection
        using_connect_as_primary_auth = False

        if jira_config.feature_flags.get("lusca-auth-always-use-connect-for-atlassian-apis-Q423"):
            logging_helper.send_to_agent_log_file("Will use connect for most API calls")
            jira_connect_or_fallback_connection = jira_atlas_connect_connection
            using_connect_as_primary_auth = True
        else:
            logging_helper.send_to_agent_log_file("Will use basic auth for most API calls")

    #######################################################################
    # Init IO Helper
    #######################################################################
    ingest_io_helper = IngestIOHelper(ingest_config=ingest_config)

    S3_OR_JELLYFISH_LOG_STATEMENT = (
        's3' if ingest_config.ingest_type == IngestionType.DIRECT_CONNECT else 'jellyfish'
    )
    if ingest_config.save_locally:
        logger.info(f"Data will be saved locally to: {ingest_io_helper.local_file_path}")
    else:
        logger.info("Data will not be saved locally")
    if ingest_config.upload_to_s3:
        logger.info(f"Data will be submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")
    else:
        logger.info(f"Data will NOT be submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")

    #######################################################################
    # Jira Projects
    #######################################################################
    with emit_event_to_jellyfish_context_manager(
        _generate_jira_event(event_name=JiraIngestEventNames.GET_JIRA_PROJECTS)
    ):
        projects_and_versions = download_projects_and_versions_and_components(
            jira_connection=jira_connect_or_fallback_connection,
            is_agent_run=ingest_config.ingest_type == IngestionType.AGENT,
            jellyfish_project_ids_to_keys=jira_config.jellyfish_project_ids_to_keys,
            jellyfish_issue_metadata=jira_config.jellyfish_issue_metadata,
            include_projects=jira_config.include_projects,
            exclude_projects=jira_config.exclude_projects,
            include_categories=jira_config.include_project_categories,
            exclude_categories=jira_config.exclude_project_categories,
        )

        project_ids = {proj["id"] for proj in projects_and_versions}
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraProjectsAndVersions.value,
            json_data=projects_and_versions,
            subdirectory=SubDirectory.JIRA,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

    if not jira_config.only_issues:
        #######################################################################
        # Jira Fields
        #######################################################################
        with emit_event_to_jellyfish_context_manager(
            _generate_jira_event(event_name=JiraIngestEventNames.GET_JIRA_FIELDS)
        ):
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraFields.value,
                subdirectory=SubDirectory.JIRA,
                json_data=download_fields(
                    jira_connect_or_fallback_connection,
                    jira_config.include_fields,
                    jira_config.exclude_fields,
                ),
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )

        ######################################################################
        # Jira Users
        #######################################################################
        with emit_event_to_jellyfish_context_manager(
            _generate_jira_event(event_name=JiraIngestEventNames.GET_JIRA_USERS)
        ):
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraUsers.value,
                subdirectory=SubDirectory.JIRA,
                json_data=download_users(
                    jira_basic_connection=jira_connect_or_fallback_connection,
                    jira_atlas_connect_connection=jira_atlas_connect_connection
                    if jira_config.should_augment_emails
                    else None,  # Use AtlasConnect for 'augment with email' subtask
                    gdpr_active=jira_config.gdpr_active,
                    search_users_by_letter_email_domain=jira_config.search_users_by_letter_email_domain,
                    required_email_domains=jira_config.required_email_domains,
                    is_email_required=jira_config.is_email_required,
                    using_connect_as_primary_auth=using_connect_as_primary_auth,
                ),
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )

        #######################################################################
        # Jira Resolutions
        #######################################################################
        with emit_event_to_jellyfish_context_manager(
            _generate_jira_event(event_name=JiraIngestEventNames.GET_JIRA_RESOLUTIONS)
        ):
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraResolutions.value,
                subdirectory=SubDirectory.JIRA,
                json_data=download_resolutions(jira_connect_or_fallback_connection),
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )

        #######################################################################
        # Jira Issue Types
        #######################################################################
        with emit_event_to_jellyfish_context_manager(
            _generate_jira_event(event_name=JiraIngestEventNames.GET_JIRA_ISSUE_TYPES)
        ):
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraIssueTypes.value,
                subdirectory=SubDirectory.JIRA,
                json_data=download_issuetypes(
                    jira_connect_or_fallback_connection, project_ids=project_ids
                ),
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )

        #######################################################################
        # Jira Link Types
        #######################################################################
        with emit_event_to_jellyfish_context_manager(
            _generate_jira_event(event_name=JiraIngestEventNames.GET_JIRA_LINK_TYPES)
        ):
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraLinkTypes.value,
                subdirectory=SubDirectory.JIRA,
                json_data=download_issuelinktypes(jira_connect_or_fallback_connection),
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )

        #######################################################################
        # Jira Priorities
        #######################################################################
        with emit_event_to_jellyfish_context_manager(
            _generate_jira_event(event_name=JiraIngestEventNames.GET_JIRA_PRIORITIES)
        ):
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraPriorities.value,
                subdirectory=SubDirectory.JIRA,
                json_data=download_priorities(jira_connect_or_fallback_connection),
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )

        #######################################################################
        # Jira Statuses
        #######################################################################
        with emit_event_to_jellyfish_context_manager(
            _generate_jira_event(event_name=JiraIngestEventNames.GET_JIRA_STATUSES)
        ):
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraStatuses.value,
                subdirectory=SubDirectory.JIRA,
                json_data=download_statuses(jira_connect_or_fallback_connection),
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )

        #######################################################################
        # Jira Boards, Sprints, and Links
        #######################################################################
        with emit_event_to_jellyfish_context_manager(
            _generate_jira_event(event_name=JiraIngestEventNames.GET_JIRA_BOARD_AND_SPRINTS)
        ):
            boards, sprints, links = download_boards_and_sprints(
                jira_connect_or_fallback_connection,
                jira_config.download_sprints,
            )
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraBoards.value,
                subdirectory=SubDirectory.JIRA,
                json_data=boards,
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )

            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraSprints.value,
                subdirectory=SubDirectory.JIRA,
                json_data=sprints,
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )

            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraBoardSprintLinks.value,
                subdirectory=SubDirectory.JIRA,
                json_data=links,
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )

    if not jira_config.skip_issues:
        # For Jira Issue operations we need to determine the batch size that
        # the JIRA provider will limit us to
        with emit_event_to_jellyfish_context_manager(
            _generate_jira_event(event_name=JiraIngestEventNames.GET_JIRA_ISSUES)
        ):
            jira_issues_batch_size = get_jira_search_batch_size(
                jira_connection=jira_connect_or_fallback_connection,
                optimistic_batch_size=jira_config.issue_batch_size,
            )

            # check if this company is on new ingest path, default to False
            if jira_config.feature_flags.get(Constants.NEW_SYNC_FF_NAME, False):
                all_downloaded_issue_ids, all_deleted_issue_ids = pull_issues_new_sync_path(
                    jira_connect_or_fallback_connection,
                    jira_config,
                    projects_and_versions,
                    jira_issues_batch_size,
                    ingest_io_helper,
                    ingest_config,
                )
            else:
                all_downloaded_issue_ids, all_deleted_issue_ids = pull_issues_old_metadata_path(
                    jira_connect_or_fallback_connection,
                    jira_issues_batch_size,
                    jira_config,
                    projects_and_versions,
                    ingest_io_helper,
                    ingest_config,
                )

            logger.info(f'{len(all_deleted_issue_ids)} issues have been detected as being deleted')
            # Write issues that got deleted
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraIssuesIdsDeleted.value,
                subdirectory=SubDirectory.JIRA,
                json_data=list(all_deleted_issue_ids),
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )

        #######################################################################
        # Jira Work Logs
        #######################################################################
        if jira_config.download_worklogs:
            with emit_event_to_jellyfish_context_manager(
                _generate_jira_event(event_name=JiraIngestEventNames.GET_JIRA_WORKLOGS)
            ):
                ingest_io_helper.write_json_to_local_or_s3(
                    object_name=JiraObject.JiraWorklogs.value,
                    subdirectory=SubDirectory.JIRA,
                    json_data=download_worklogs(
                        jira_connect_or_fallback_connection,
                        all_downloaded_issue_ids,
                        jira_config.work_logs_pull_from,
                    ),
                    save_locally=ingest_config.save_locally,
                    upload_to_s3=ingest_config.upload_to_s3,
                )
    else:
        logger.info(
            f"Skipping issues and worklogs bc config.skip_issues is {jira_config.skip_issues}"
        )

    if ingest_config.save_locally:
        logger.info(f"Data has been saved locally to: {ingest_io_helper.local_file_path}")
    else:
        logger.info(
            f"Data has not been saved locally, because save_locally was set to false in the ingest config!"
        )

    if ingest_config.upload_to_s3:
        logger.info(f"Data has been submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")
    else:
        logger.info(f"Data was not submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")

    return True



from unittest.mock import patch

import pytest
from jira import JIRAError

from jf_ingest.config import JiraAuthConfig
from jf_ingest.jf_jira.auth import JiraAuthenticationException, JiraAuthMethod, get_jira_connection
from jf_ingest.jf_jira.exceptions import JiraRetryLimitExceeded


def _get_jira_auth_config(args):
    company_slug = "test_company"
    url = "https://test-co.atlassian.net/"
    return JiraAuthConfig(company_slug=company_slug, url=url, **args)


def test_jira_conn_basic_auth_no_auth():
    config = _get_jira_auth_config({"available_auth_methods": [JiraAuthMethod.BasicAuth]})

    with pytest.raises(RuntimeError) as excinfo:
        get_jira_connection(config)

    assert (
        str(excinfo.value)
        == f"No valid basic authentication mechanism for {config.url} - need a username/password combo or a personal access token"
    )


def test_jira_conn_basic_auth_user_password():
    config = _get_jira_auth_config(
        {
            "available_auth_methods": [JiraAuthMethod.BasicAuth],
            "user": "jira_user@test.co",
            "password": "test_password",  # pragma: allowlist secret
        }
    )

    # TEST BASIC AUTH ERROR LOGIC
    with patch("jf_ingest.jf_jira.auth.JIRA", side_effect=JIRAError(status_code=401)):
        with pytest.raises(JiraAuthenticationException):
            get_jira_connection(config, max_retries=0)

    # TEST BASIC AUTH ERROR LOGIC
    with patch("jf_ingest.jf_jira.auth.JIRA", side_effect=JIRAError(status_code=403)):
        with pytest.raises(JiraAuthenticationException):
            get_jira_connection(config, max_retries=0)


def test_jira_conn_basic_auth_personal_access_token():
    config = _get_jira_auth_config(
        {
            "available_auth_methods": [JiraAuthMethod.BasicAuth],
            "personal_access_token": "a_PAT",
        }
    )

    # TEST BASIC AUTH ERROR LOGIC
    with patch("jf_ingest.jf_jira.auth.JIRA", side_effect=JIRAError(status_code=401)):
        with pytest.raises(JiraAuthenticationException):
            get_jira_connection(config, max_retries=0)

    # TEST BASIC AUTH ERROR LOGIC
    with patch("jf_ingest.jf_jira.auth.JIRA", side_effect=JIRAError(status_code=403)):
        with pytest.raises(JiraAuthenticationException):
            get_jira_connection(config, max_retries=0)



import logging
import re
from datetime import datetime, timedelta, timezone
from itertools import chain
from typing import Any, Dict, Generator, List, Optional

from dateutil import parser

from jf_ingest import logging_helper
from jf_ingest.config import GitConfig, IngestionType
from jf_ingest.jf_git.adapters import GitAdapter
from jf_ingest.jf_git.clients.azure_devops import AzureDevopsClient
from jf_ingest.jf_git.standardized_models import (
    PullRequestReviewState,
    StandardizedBranch,
    StandardizedCommit,
    StandardizedFileData,
    StandardizedLabel,
    StandardizedOrganization,
    StandardizedPullRequest,
    StandardizedPullRequestComment,
    StandardizedPullRequestMetadata,
    StandardizedPullRequestReview,
    StandardizedRepository,
    StandardizedTeam,
    StandardizedUser,
)
from jf_ingest.utils import hash_filename


class AzureDevopsAdapter(GitAdapter):
    def __init__(self, config: GitConfig):
        # Git Config options
        self.client = AzureDevopsClient(config.git_auth_config)
        self.config = config
        self.repo_id_to_name_lookup: Dict = {}
        self.repo_id_to_project_id: Dict = {}

    @property
    def supports_get_all_users(self) -> bool:
        """
        mpk 5/12/23: Azure DevOps Server 2020 (on-prem, encountered for at least one client) does NOT support a
        get_all_users method in the API. Cloud-based Azure DevOps Services (i.e., dev.azure.com) does. We
        haven't yet encountered other ADO server versions (like ADO Server 2022), so don't know whether or
        not those other versions will support get_all_users. For now I'll just assume that anything that
        runs the API version "6.1*" (like ADO Server 2020) will NOT support get_all_users.

        Returns:
            bool: True if we are in a version that is not 6.1
        """
        return not self.client.api_version.startswith('6.1')

    def _project_name_from_repo(self, standardized_repo: StandardizedRepository) -> str:
        '''
        The Project Name should be pre-loaded into the self.repo_id_to_project_id object,
        which gets set in the adapter get_repos function, and mapped to the repo ID. If
        it isn't, call the get_all_repos() client function again and save the results
        to the adapter.

        Returns the project name of the provided standardized_repo
        '''
        if project_name := self.repo_id_to_project_id.get(standardized_repo.id):
            return project_name
        else:
            logging_helper.send_to_agent_log_file(
                f'Repos were not preloaded? Loading them now', level=logging.WARNING
            )
            for repo in self.client.get_all_repos():
                self.repo_id_to_project_id[repo['id']] = repo['project']['name']

        return self.repo_id_to_project_id[standardized_repo.id]

    def get_organizations(self) -> List[StandardizedOrganization]:
        # I think Azure devops can only have one org per instance??
        org = self.config.git_organizations[0]
        return [
            _standardize_organization(
                org, self.client.base_url, self.config.git_redact_names_and_urls
            )
        ]

    def get_api_scopes(self) -> str:
        raise NotImplementedError()

    def get_users(
        self, standardized_org: StandardizedOrganization, limit: int = None
    ) -> Generator[StandardizedUser, None, None]:
        i = 1
        for api_user in self.client.get_graph_users(org_name=standardized_org.login):
            standardized_user = _standardize_graph_user(api_user)
            if standardized_user:
                yield standardized_user
                if limit and i >= limit:
                    return
                i += 1

    def get_teams(
        self, standardized_organization: StandardizedOrganization, limit: int = None
    ) -> Generator[StandardizedTeam, None, None]:
        for i, api_team in enumerate(
            self.client.get_teams(org_name=standardized_organization.login), start=1
        ):
            api_team_members = self.client.get_team_users(
                org_name=standardized_organization.login, team_descriptor=api_team['descriptor']
            )
            yield _standardize_team(api_team, api_team_members=api_team_members)
            if limit and i >= limit:
                return

    def get_repos(
        self, standardized_organization: StandardizedOrganization
    ) -> Generator[StandardizedRepository, None, None]:
        for api_repo in self.client.get_all_repos(org_name=standardized_organization.login):
            self.repo_id_to_project_id[api_repo['id']] = api_repo['project']['name']
            yield _standardize_repo(
                org=standardized_organization,
                api_repo=api_repo,
                redact_names_and_urls=self.config.git_redact_names_and_urls,
            )

    def get_commits_for_default_branch(
        self,
        standardized_repo: StandardizedRepository,
        limit: int = None,
        pull_since: datetime = None,
        pull_until: datetime = None,
    ) -> Generator[StandardizedCommit, None, None]:
        for i, api_commit in enumerate(
            self.client.get_commits(
                org_name=standardized_repo.organization.login,
                project_name=self._project_name_from_repo(standardized_repo),
                repo_id=standardized_repo.id,
                branch_name=standardized_repo.default_branch_name,
                from_date=pull_since,
                to_date=pull_until,
            ),
            start=1,
        ):
            yield _standardize_commit(
                api_commit,
                branch_name=standardized_repo.default_branch_name,
                standardized_repo=standardized_repo,
                strip_text_content=self.config.git_strip_text_content,
                redact_names_and_urls=self.config.git_redact_names_and_urls,
            )
            if limit and i >= limit:
                return

    def get_branches_for_repo(
        self,
        standardized_repo: StandardizedRepository,
        pull_branches: Optional[bool] = False,
    ) -> Generator[StandardizedBranch, None, None]:
        if pull_branches:
            for api_branch in self.client.get_branches(
                org_name=standardized_repo.organization.login,
                project_name=self._project_name_from_repo(standardized_repo),
                repo_id=standardized_repo.id,
            ):
                yield _standardize_branch(
                    api_branch=api_branch,
                    standardized_repo=standardized_repo,
                    redact_names_and_urls=self.config.git_redact_names_and_urls,
                )
        else:
            # Above, if we're pulling all branches, it's safe to assume that the default branch
            # will be included in that
            # When we don't pull all branches, always return default branch
            if standardized_repo.default_branch_name:
                yield StandardizedBranch(
                    repo_id=standardized_repo.id,
                    name=standardized_repo.default_branch_name,
                    sha=standardized_repo.default_branch_sha,
                    is_default=True,
                )

    def get_commits_for_branches(
        self,
        standardized_repo: StandardizedRepository,
        branches: List[StandardizedBranch],
        pull_since: datetime = None,
        pull_until: datetime = None,
    ) -> Generator[StandardizedCommit, None, None]:
        for branch_name in self.get_filtered_branches(standardized_repo, branches):
            try:
                login = standardized_repo.organization.login
                for api_commit in self.client.get_commits(
                    org_name=login,
                    project_name=self._project_name_from_repo(standardized_repo),
                    repo_id=standardized_repo.id,
                    branch_name=branch_name,
                    from_date=pull_since,
                    to_date=pull_until,
                ):
                    yield _standardize_commit(
                        api_commit=api_commit,
                        branch_name=branch_name,
                        standardized_repo=standardized_repo,
                        strip_text_content=self.config.git_strip_text_content,
                        redact_names_and_urls=self.config.git_redact_names_and_urls,
                    )

            except Exception as e:
                logging_helper.send_to_agent_log_file(
                    f'Got exception for branch {branch_name}: {e}. Skipping...',
                    level=logging.WARNING,
                )

    def get_pr_updated_date(
        self, standardized_repo: StandardizedRepository, api_pr: Dict
    ) -> datetime:
        if 'closedDate' in api_pr:
            return parser.parse(api_pr['closedDate'])

        org_name = standardized_repo.organization.login
        project_name = self._project_name_from_repo(standardized_repo)
        repo_id = standardized_repo.id
        pr_iterations = self.client.get_pull_request_iterations(
            org_name=org_name,
            project_name=project_name,
            repo_id=repo_id,
            pr_id=api_pr['pullRequestId'],
        )
        assert len(pr_iterations) > 0
        # Get the latest iteration and grab the updatedDate value
        return parser.parse(pr_iterations[-1]['updatedDate'])

    def get_pr_metadata(
        self,
        standardized_repo: StandardizedRepository,
        limit: int = None,
    ) -> Generator[StandardizedPullRequestMetadata, None, None]:
        supports_pr_filter_func = self.git_provider_pr_endpoint_supports_date_filtering
        raise NotImplemented(
            'This function is NOT implemented, because ADO supports PR date filtering '
            f'({supports_pr_filter_func.__name__} is {supports_pr_filter_func()})'
        )

    def git_provider_pr_endpoint_supports_date_filtering(self):
        return True

    def _get_prs_helper(
        self,
        standardized_repo: StandardizedRepository,
        start_window: datetime,
        end_window: datetime,
    ) -> Generator[Dict, None, None]:
        prs: List[Dict] = []
        tomorrow = datetime.now().astimezone(timezone.utc) + timedelta(days=1)

        def _get_prs_from_client_wrapper(_end_window: datetime, status: str, filter_by: str):
            return self.client.get_pull_requests(
                org_name=standardized_repo.organization.login,
                project_name=self._project_name_from_repo(standardized_repo=standardized_repo),
                repo_id=standardized_repo.id,
                start_window=start_window,
                end_window=_end_window,
                status=status,
                filter_by=filter_by,
            )

        abandoned_prs_generator = _get_prs_from_client_wrapper(end_window, 'abandoned', 'closed')
        completed_prs_generator = _get_prs_from_client_wrapper(end_window, 'completed', 'closed')
        active_prs = _get_prs_from_client_wrapper(tomorrow, 'active', 'open')
        return chain(abandoned_prs_generator, completed_prs_generator, active_prs)

    def get_prs(
        self,
        standardized_repo: StandardizedRepository,
        pull_files_for_pr: bool = False,
        hash_files_for_prs: bool = False,
        limit: int = None,  # Not used by ADO adapter
        start_cursor: Any = 0,
        start_window: datetime = None,
        end_window: datetime = None,
    ) -> Generator[StandardizedPullRequest, None, None]:
        try:
            login = standardized_repo.organization.login
            repo_id = standardized_repo.id
            project_name = self._project_name_from_repo(standardized_repo)

            api_prs = self._get_prs_helper(
                standardized_repo=standardized_repo,
                start_window=start_window,
                end_window=end_window,
            )

            for i, api_pr in enumerate(
                api_prs,
                start=1,
            ):
                try:
                    pr_id = api_pr['pullRequestId']
                    head_branch = api_pr['sourceRefName'].replace('refs/heads/', '')
                    base_branch = api_pr['targetRefName'].replace('refs/heads/', '')
                    files = []
                    if 'lastMergeTargetCommit' in api_pr and 'lastMergeCommit' in api_pr:
                        change_counts = self.client.get_pull_request_changes_counts(
                            org_name=login,
                            repo_id=standardized_repo.id,
                            base_sha=api_pr['lastMergeTargetCommit']['commitId'],
                            target_sha=api_pr['lastMergeCommit']['commitId'],
                        )
                        changed_count = sum(change_counts.values())
                        additions = change_counts.get('Add', 0)
                        deletions = change_counts.get('Delete', 0)
                        if pull_files_for_pr:
                            changes = self.client.get_pull_request_diff(
                                org_name=login,
                                repo_id=standardized_repo.id,
                                base_sha=api_pr['lastMergeTargetCommit']['commitId'],
                                target_sha=api_pr['lastMergeCommit']['commitId'],
                            )
                            files = [c['item']['path'] for c in changes]
                            if hash_files_for_prs:
                                files = [hash_filename(file) for file in files]
                        else:
                            files = []

                    pr_updated_at = self.get_pr_updated_date(
                        standardized_repo=standardized_repo, api_pr=api_pr
                    )

                    commits = [
                        _standardize_commit(
                            api_commit=api_commit,
                            branch_name=head_branch,
                            standardized_repo=standardized_repo,
                            strip_text_content=self.config.git_strip_text_content,
                            redact_names_and_urls=self.config.git_redact_names_and_urls,
                        )
                        for api_commit in self.client.get_pr_commits(
                            org_name=login,
                            project_name=project_name,
                            repo_id=repo_id,
                            pr_id=pr_id,
                        )
                    ]
                    # If a merge commit is in here, retroactively mark one of the commits
                    # as a merge commit
                    if api_pr['status'] == 'completed' and 'lastMergeCommit' in api_pr:
                        for commit in commits:
                            if commit.hash == api_pr['lastMergeCommit']['commitId']:
                                commit.is_merge = True

                    comments: List[StandardizedPullRequestComment] = []
                    reviews: List[StandardizedPullRequestReview] = []
                    for comment_thread in self.client.get_pr_comment_threads(
                        org_name=login, project_name=project_name, repo_id=repo_id, pr_id=pr_id
                    ):
                        comment_thread_id = comment_thread['id']
                        for api_comment in comment_thread['comments']:
                            comments.append(
                                _standardize_pull_request_comment(
                                    api_comment, self.config.git_strip_text_content
                                )
                            )
                            pr_review = _standardize_pull_request_review(
                                api_comment=api_comment,
                                pr_id=pr_id,
                                comment_thread_id=comment_thread_id,
                            )

                            if pr_review:
                                reviews.append(pr_review)

                    labels = [label['name'] for label in api_pr.get('labels', [])]

                    yield _standardize_pull_request(
                        api_pr,
                        standardized_repo,
                        pr_updated_at=pr_updated_at,
                        commits=commits,
                        comments=comments,
                        approvals=reviews,
                        labels=labels,
                        additions=additions,
                        deletions=deletions,
                        changed_count=changed_count,
                        files=files,
                        strip_text_content=self.config.git_strip_text_content,
                        redact_names_and_urls=self.config.git_redact_names_and_urls,
                    )
                    if limit and i >= limit:
                        return
                except Exception:
                    # if something goes wrong with normalizing one of the prs - don't stop pulling. try
                    # the next one.
                    pr_id = api_pr.get('id', '?')
                    logging_helper.send_to_agent_log_file(
                        f'normalizing PR {pr_id} from repo {standardized_repo.name} ({standardized_repo.id}). Skipping...',
                        level=logging.WARNING,
                        exc_info=True,
                    )

        except Exception:
            logging_helper.send_to_agent_log_file(
                f'normalizing PRs from repo {standardized_repo.name} ({standardized_repo.id}). Skipping...',
                level=logging.WARNING,
                exc_info=True,
            )


'''

Massage Functions (standardize raw JSON data to StandardizedPython Models)

'''


def _standardize_graph_user(api_user: Dict) -> StandardizedUser:
    """Standardize the raw GraphUser type to our StandardizedUser.
    This is for users returned from the "Graph" API.
    https://learn.microsoft.com/en-us/rest/api/azure/devops/graph/users/list?view=azure-devops-rest-7.1&tabs=HTTP#graphuser

    Args:
        api_user (Dict): A raw dictionary from the ADO API representing a GraphUser

    Returns:
        StandardizedUser: A standardized User object
    """
    if not (user_id := api_user.get('descriptor')):
        return None
    return StandardizedUser(
        id=user_id,
        login=api_user['principalName'],
        name=api_user['displayName'],
        email=api_user['mailAddress'],
        url=api_user['url'],
    )


def _standardize_team(api_team: Dict, api_team_members: List[Dict]) -> StandardizedTeam:
    members = []
    for member in api_team_members:
        if standardized_member := _standardize_graph_user(member):
            members.append(standardized_member)

    return StandardizedTeam(
        id=api_team['descriptor'],
        slug=api_team['descriptor'],
        name=api_team['displayName'],
        description=api_team['description'],
        members=members,
    )


def _standardize_repo(
    org: StandardizedOrganization, api_repo: dict, redact_names_and_urls: bool
) -> StandardizedRepository:
    # Sanitize Repo Name
    repo_name = (
        api_repo['name']
        if not redact_names_and_urls
        else GitAdapter.repo_redactor.redact_name(api_repo['name'])
    )

    # Sanitize URL
    url = api_repo['url'] if not redact_names_and_urls else None

    # Sanitize Branch Name
    if 'defaultBranch' in api_repo:
        default_branch_name = api_repo['defaultBranch'].replace('refs/heads/', '')
        default_branch_name = (
            default_branch_name
            if not redact_names_and_urls
            else GitAdapter.branch_redactor.redact_name(default_branch_name)
        )
    else:
        default_branch_name = None

    return StandardizedRepository(
        id=api_repo['id'],
        name=repo_name,
        full_name=f'{org.login}/{repo_name}',
        url=url,
        is_fork=api_repo.get('isFork', False),
        organization=org,
        default_branch_name=default_branch_name,
        default_branch_sha=None,  # TODO: looks like this isn't part of the base object, is it important? Do we NEED it?
    )


def _standardize_branch(
    api_branch: dict, standardized_repo: StandardizedRepository, redact_names_and_urls: bool
) -> StandardizedBranch:
    if not api_branch:
        return None
    if not api_branch.get('name'):
        return None

    branch_name = api_branch['name'].replace('refs/heads/', '')
    return StandardizedBranch(
        repo_id=standardized_repo.id,
        name=branch_name
        if not redact_names_and_urls
        else GitAdapter.branch_redactor.redact_name(branch_name),
        sha=api_branch['objectId'],
        is_default=branch_name == standardized_repo.default_branch_name,
    )


def _standardize_git_user_data(user: Dict) -> StandardizedUser:
    """Standardize the GitUserDate type, which is the type of user returned
    on commits.
    https://learn.microsoft.com/en-us/rest/api/azure/devops/git/commits/get-commits?view=azure-devops-rest-7.1&tabs=HTTP#gituserdate

    Args:
        user (Dict): A GitUserDict returned from the API

    Returns:
        StandardizedUser: The Standardized User
    """
    if not (email := user.get('email')):
        return None
    return StandardizedUser(
        id=email,
        name=user.get('name'),
        login=email,
        email=email,
        url=None,
        account_id=None,
    )


def _standardize_organization(
    org_login: str, url: str, redact_names_and_urls: bool
) -> StandardizedOrganization:
    # TODO this is probably wrong
    return StandardizedOrganization(
        id=org_login,
        name=org_login
        if not redact_names_and_urls
        else GitAdapter.organization_redactor.redact_name(org_login),
        login=org_login,
        url=url if not redact_names_and_urls else None,
    )


def _standardize_commit(
    api_commit: dict,
    branch_name: str,
    standardized_repo: StandardizedRepository,
    strip_text_content: bool,
    redact_names_and_urls: bool,
) -> StandardizedCommit:
    # NOTE: The commit object only returns "GitUserDate" types
    author = _standardize_git_user_data(api_commit['author'])
    return StandardizedCommit(
        hash=api_commit['commitId'],
        url=api_commit['url'] if not redact_names_and_urls else None,
        message=GitAdapter.sanitize_text(api_commit['comment'], strip_text_content),
        branch_name=branch_name
        if not redact_names_and_urls
        else GitAdapter.branch_redactor.redact_name(branch_name),
        commit_date=parser.parse(api_commit['committer']['date']),
        author_date=parser.parse(api_commit['author']['date']),
        author=author if author else None,
        repo=standardized_repo.short(),
        # XXX: The commit we get back from the API doesn't include the "parents" item; the docs seem
        # to indicate it's supposed to. Without having the "parents" item, we can't tell whether or
        # not the commit is a merge commit. I think we could get the "parents" item if we use the
        # "Get" API to get a single commit rather than the "GetCommits" API to get a list of commits.
        # But that'd cause us to make way more API requests. For now at least let's avoid that and
        # just set is_merge to False (perhaps erroneously) for all commits.
        # NOTE (update): For Pull Requests we WILL mark merge commits properly as part of PR Ingestion.
        # In theory all, if not most, merge commits will come to us via PRs, so we should eventaully
        # mark all merge commits with this strategy
        is_merge=False,
    )


def _standardize_identity_ref_user(user: Dict) -> StandardizedUser:
    """Standardize the ADO IdentityRef User object.
    This is the type commonly returned by the PR related data objects.
    https://learn.microsoft.com/en-us/rest/api/azure/devops/git/pull-request-threads/list?view=azure-devops-rest-7.1&tabs=HTTP#identityref

    Args:
        user (Dict): A raw "IdentityRef" object type from ADO

    Returns:
        StandardizedUser: A standardized User Object
    """
    user_id = user.get('descriptor') or user.get('uniqueName') or user.get('id')
    if not user_id:
        return None
    return StandardizedUser(
        id=user_id,
        name=user['displayName'],
        login=user_id,
        email=user.get('uniqueName'),
        url=user['url'],
        account_id=user['id'],
    )


def _standardize_pull_request_review(
    api_comment: Dict, pr_id: int, comment_thread_id: int
) -> StandardizedPullRequestReview:
    """Normalize an API comment into a review.
    State is derived by checking if this is a system comment type, and by
    checking the comment content.
    User is derived from the comment dictionary
    foreign-id is derived from Pull Request ID, Comment Thread ID, and Comment ID. Honestly,
    this probably isn't that useful but at least it's something

    Args:
        api_comment (Dict): Used to derive approval state and commenter
        pr_id (int): Used as part of the Foreign ID field
        comment_thread_id (int): Used as part of the Foreign ID field

    Returns:
        StandardizedPullRequestReview: A standardized Pull Request Review
    """
    api_comment_author = api_comment['author']
    commenter = (
        _standardize_identity_ref_user(api_comment_author)
        if 'uniqueName' in api_comment_author
        else None
    )

    comment_type = api_comment.get('commentType')
    comment_content = api_comment.get('content')

    # If the comment is an approval, construct a StandardizedPullRequestApproval
    review_state: PullRequestReviewState = None
    if comment_type == 'system' and 'voted 5' in comment_content:
        # This is "approved the pull request with suggestions"
        review_state = PullRequestReviewState.APPROVED
    elif comment_type == 'system' and 'voted 10' in comment_content:
        # This is "approved the pull request"
        review_state = PullRequestReviewState.APPROVED
    elif comment_type == 'system' and 'voted -10' in comment_content:
        # This is "rejected the pull request"
        review_state = PullRequestReviewState.DISMISSED
    elif comment_type == 'system' and 'voted -5' in comment_content:
        # This is "is waiting for the author"
        review_state = PullRequestReviewState.CHANGES_REQUESTED
    elif comment_type == 'text':
        review_state = PullRequestReviewState.COMMENTED

    if review_state:
        return StandardizedPullRequestReview(
            user=commenter,
            foreign_id=f"{pr_id}-{comment_thread_id}-{api_comment['id']}",
            review_state=review_state.name,
        )


def _standardize_pull_request_comment(
    api_comment: Dict, strip_text_content: bool
) -> StandardizedPullRequestComment:
    api_comment_author = api_comment['author']
    commenter = (
        _standardize_identity_ref_user(api_comment_author)
        if 'uniqueName' in api_comment_author
        else None
    )
    comment_body = api_comment.get('content', '')
    return StandardizedPullRequestComment(
        user=commenter,
        body=GitAdapter.sanitize_text(comment_body, strip_text_content),
        created_at=parser.parse(api_comment['publishedDate']),
        system_generated=api_comment.get('commentType') == 'system',
    )


def _standardize_pull_request(
    api_pr: Dict,
    standardized_repo: StandardizedRepository,
    pr_updated_at: datetime,
    commits: List[StandardizedCommit],
    comments: List[StandardizedPullRequestComment],
    approvals: List[StandardizedPullRequestReview],
    labels: List[StandardizedLabel],
    additions: int,
    deletions: int,
    changed_count: int,
    files: List[StandardizedFileData],
    strip_text_content: bool,
    redact_names_and_urls: bool,
) -> StandardizedPullRequest:
    #
    # PR Status
    pr_status = api_pr['status']
    is_merged = pr_status == 'completed'
    is_closed = pr_status != 'active'
    closed_date = parser.parse(api_pr['closedDate']) if 'closedDate' in api_pr else None
    merge_date = closed_date if is_merged else None

    #
    # User Transformation
    author = _standardize_identity_ref_user(api_pr['createdBy'])

    merge_commits = [c for c in commits if c.is_merge]
    merge_commit = merge_commits[0] if len(merge_commits) == 1 else None
    if merge_commit:
        merged_by = merge_commit.author
    else:
        merged_by = None

    title = GitAdapter.sanitize_text(api_pr['title'], strip_text_content)
    body = GitAdapter.sanitize_text(api_pr.get('description', ''), strip_text_content)

    base_branch_name = api_pr['targetRefName'].replace('refs/heads/', '')
    head_branch_name = api_pr['sourceRefName'].replace('refs/heads/', '')
    base_branch = (
        base_branch_name
        if not redact_names_and_urls
        else GitAdapter.branch_redactor.redact_name(base_branch_name)
    )
    head_branch = (
        head_branch_name
        if not redact_names_and_urls
        else GitAdapter.branch_redactor.redact_name(head_branch_name)
    )

    return StandardizedPullRequest(
        id=api_pr['pullRequestId'],
        additions=additions,
        deletions=deletions,
        changed_files=changed_count,
        is_closed=is_closed,
        is_merged=is_merged,
        created_at=parser.parse(api_pr['creationDate']),
        updated_at=pr_updated_at,
        merge_date=merge_date,
        closed_date=closed_date,
        title=title,
        body=body,
        url=api_pr['url'] if not redact_names_and_urls else None,
        base_branch=base_branch,
        head_branch=head_branch,
        author=author if author else None,
        merged_by=merged_by if merged_by else None,
        commits=commits,
        merge_commit=merge_commit,
        comments=comments,
        approvals=approvals,
        base_repo=standardized_repo,
        head_repo=standardized_repo,
        labels=labels,
        files=files,
    )



import json

import requests_mock

from jf_ingest.jf_jira.downloaders import download_issuetypes
from tests.jf_jira.utils import _register_jira_uri, get_fixture_file_data, get_jira_mock_connection

ISSUE_TYPE_SORT_KEY = lambda x: x["id"]


def test_download_issuetypes_no_projects():
    with requests_mock.Mocker() as m:
        issuetype_data_str = get_fixture_file_data("api_responses/issuetype.json")
        _register_jira_uri(m, endpoint="issuetype", return_value=issuetype_data_str)

        issuetypes = download_issuetypes(get_jira_mock_connection(), project_ids=[])

        issuetype_data = json.loads(issuetype_data_str)
        issuetypes_with_no_project_scope = [
            issuetype for issuetype in issuetype_data if "scope" not in issuetype
        ]

        assert type(issuetypes) == list
        assert len(issuetypes) > 0
        assert issuetypes.sort(key=ISSUE_TYPE_SORT_KEY) == issuetypes_with_no_project_scope.sort(
            key=ISSUE_TYPE_SORT_KEY
        )


def test_download_issuetypes_with_projects():
    with requests_mock.Mocker() as m:
        issuetype_data_str = get_fixture_file_data("api_responses/issuetype.json")
        _register_jira_uri(m, endpoint="issuetype", return_value=issuetype_data_str)

        projects_to_include = ["10022", "10009"]

        issuetype_data = json.loads(issuetype_data_str)

        # Issue types we want
        issuetypes_with_no_project_scope = [
            issuetype for issuetype in issuetype_data if "scope" not in issuetype
        ]
        issue_types_with_included_projects = [
            issuetype
            for issuetype in issuetype_data
            if "scope" in issuetype
            and issuetype["scope"]["type"] == "PROJECT"
            and issuetype["scope"]["project"]["id"] in projects_to_include
        ]
        issuetypes_we_want = [
            *issue_types_with_included_projects,
            *issuetypes_with_no_project_scope,
        ]

        # Issue types we don't want
        issuetypes_we_do_not_want = [
            issuetype
            for issuetype in issuetype_data
            if "scope" in issuetype
            and issuetype["scope"]["type"] == "PROJECT"
            and issuetype["scope"]["project"]["id"] not in projects_to_include
        ]

        issuetypes = download_issuetypes(
            get_jira_mock_connection(), project_ids=projects_to_include
        )
        assert len(issuetypes) > 0
        assert issuetypes.sort(key=ISSUE_TYPE_SORT_KEY) == issuetypes_we_want.sort(
            key=ISSUE_TYPE_SORT_KEY
        )
        for issuetype in issuetypes:
            assert issuetype not in issuetypes_we_do_not_want



from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum
from typing import Dict, List


@dataclass
class StandardizedUser:
    id: str
    name: str
    login: str
    email: str = None
    url: str = None
    account_id: str = None


@dataclass
class StandardizedTeam:
    id: str
    slug: str
    name: str
    description: str
    members: list[StandardizedUser]


@dataclass
class StandardizedBranch:
    name: str
    sha: str
    repo_id: str
    is_default: bool


@dataclass
class StandardizedOrganization:
    id: str
    name: str
    login: str
    url: str


@dataclass
class StandardizedShortRepository:
    id: int
    name: str
    url: str


@dataclass
class StandardizedRepository:
    id: str
    name: str
    full_name: str
    url: str
    is_fork: bool
    default_branch_name: str
    default_branch_sha: str
    organization: StandardizedOrganization
    commits_backpopulated_to: datetime = None
    prs_backpopulated_to: datetime = None

    def short(self):
        # return the short form of Standardized Repository
        return StandardizedShortRepository(id=self.id, name=self.name, url=self.url)


@dataclass
class StandardizedCommit:
    hash: str
    url: str
    message: str
    commit_date: datetime
    author_date: datetime
    author: StandardizedUser
    repo: StandardizedShortRepository
    is_merge: bool
    branch_name: str = None


@dataclass
class StandardizedPullRequestComment:
    user: StandardizedUser
    body: str
    created_at: str
    system_generated: bool = None


# TODO: This exists in the Jellyfish source code
# and has been copied over. The source code should be
# replaced with this, so we have one "source of truth".
# NOTE: These enum string values are based off of Github,
# we need to normalize all other providers to this
class PullRequestReviewState(IntEnum):
    UNKNOWN = 0
    PENDING = 1
    APPROVED = 2
    COMMENTED = 3
    CHANGES_REQUESTED = 4
    DISMISSED = 5


@dataclass
class StandardizedPullRequestReview:
    user: StandardizedUser
    foreign_id: int
    review_state: str


@dataclass
class StandardizedLabel:
    id: int
    name: str
    default: bool
    description: str


@dataclass
class StandardizedFileData:
    status: str
    changes: int
    additions: int
    deletions: int


@dataclass
class StandardizedPullRequest:
    id: any
    additions: int
    deletions: int
    changed_files: int
    is_closed: bool
    is_merged: bool
    created_at: str
    updated_at: datetime
    merge_date: datetime
    closed_date: datetime
    title: str
    body: str
    url: str
    base_branch: str
    head_branch: str
    author: StandardizedUser
    merged_by: StandardizedUser
    commits: List[StandardizedCommit]
    merge_commit: StandardizedCommit
    comments: List[StandardizedPullRequestComment]
    approvals: List[StandardizedPullRequestReview]
    base_repo: StandardizedShortRepository
    head_repo: StandardizedShortRepository
    labels: List[StandardizedLabel]
    files: Dict[str, StandardizedFileData]


@dataclass
class StandardizedPullRequestMetadata:
    id: any
    updated_at: datetime
    # API Index place holder, needed if the adapter DOES NOT support PullRequest API Time filtering
    # (git_provider_pr_endpoint_supports_date_filtering)
    api_index: any = None



import fnmatch
import logging
import re
from abc import ABC, abstractmethod
from dataclasses import asdict
from datetime import date, datetime, timedelta, timezone
from enum import Enum
from itertools import chain
from typing import Any, Dict, Generator, List, NamedTuple, Optional

from jf_ingest import diagnostics, logging_helper
from jf_ingest.config import GitConfig, GitProvider, GitProviderInJellyfishRepo, IngestionConfig
from jf_ingest.constants import Constants
from jf_ingest.events.models import GitIngestEvent, GitIngestionEventNames, IngestType
from jf_ingest.events.utils import emit_event_to_jellyfish_context_manager
from jf_ingest.file_operations import IngestIOHelper, SubDirectory
from jf_ingest.jf_git.exceptions import GitProviderUnavailable
from jf_ingest.jf_git.standardized_models import (
    StandardizedBranch,
    StandardizedCommit,
    StandardizedOrganization,
    StandardizedPullRequest,
    StandardizedPullRequestMetadata,
    StandardizedRepository,
    StandardizedTeam,
    StandardizedUser,
)
from jf_ingest.name_redactor import NameRedactor
from jf_ingest.utils import (
    batch_iterable,
    batch_iterable_by_bytes_size,
    get_jellyfish_company_slug,
    init_jf_ingest_run,
    tqdm_to_logger,
)

logger = logging.getLogger(__name__)

'''

    Constants

'''
# NOTE: ONLY GITHUB IS CURRENTLY SUPPORTED!!!!
BBS_PROVIDER = 'bitbucket_server'
BBC_PROVIDER = 'bitbucket_cloud'
GH_PROVIDER = 'github'
GL_PROVIDER = 'gitlab'
PROVIDERS = [GL_PROVIDER, GH_PROVIDER, BBS_PROVIDER, BBC_PROVIDER]


class BackpopulationWindow(NamedTuple):
    backpopulation_window_start: datetime
    backpopulation_window_end: datetime


def _transform_dataclass_list_to_dict(dataclass_objects: List[Any]) -> List[Dict]:
    """Helper function for taking a list of objects that inherit from Dataclass and
    transforming them to a list of dictionary objects

    Args:
        dataclass_objects (List[DataclassInstance]): A list of Dataclass Instances

    Returns:
        List[Dict]: A list of dictionaries
    """
    return [asdict(dc_object) for dc_object in dataclass_objects]


class GitObject(Enum):
    GitOrganizations = "git_data_organizations"
    GitUsers = "git_data_users"
    GitTeams = "git_data_teams"
    GitRepositories = "git_data_repos"
    GitBranches = "git_data_branches"
    GitCommits = "git_data_commits"
    GitPullRequests = "git_data_prs"


def _generate_git_ingest_event(event_name: str, git_config: GitConfig) -> GitIngestEvent:
    return GitIngestEvent(
        company_slug=get_jellyfish_company_slug(),
        ingest_type=IngestType.GIT,
        event_name=event_name,
        git_instance=git_config.instance_slug,
        git_provider=git_config.git_provider.value,
    )


class GitAdapter(ABC):
    config: GitConfig
    PULL_REQUEST_BATCH_SIZE_IN_BYTES = (
        50 * Constants.MB_SIZE_IN_BYTES
    )  # PRs can be huge and of variable size. We need to limit them by batch size in bytes
    NUMBER_OF_COMMITS_PER_BATCH = (
        30000  # Commits are generally uniform in size. This is ~50 MBs per commit batch
    )

    branch_redactor = NameRedactor(preserve_names=['master', 'develop'])
    organization_redactor = NameRedactor()
    repo_redactor = NameRedactor()

    @staticmethod
    def sanitize_text(text: str, strip_text_content: bool) -> str:
        """Helper function for removing Jira keys out of commit
        messages, PR bodies, and other git text bodies that might
        contain sensitive Jira Keys

        Args:
            text (str): A given string to "sanitize"
            strip_text_content (bool): A boolean of whether we should sanitize it or not

        Returns:
            str: The inputted text str, but with any Jira Key purged out of it
        """
        # NOTE: This module is used only by git, but we need to clean up Jira
        # keys out of commit messages. That's what this regex is for
        JIRA_KEY_REGEX = re.compile(r'([a-z0-9]+)[-|_|/| ]?(\d+)', re.IGNORECASE)

        if not text or not strip_text_content:
            return text

        regex_matches: List[str] = JIRA_KEY_REGEX.findall(text)

        return (' ').join(
            {f'{match[0].upper().strip()}-{match[1].upper().strip()}' for match in regex_matches}
        )

    @staticmethod
    def get_git_adapter(config: GitConfig):
        """Static function for generating a GitAdapter from a provided GitConfig object

        Args:
            config (GitConfig): A git configuration data object. The specific GitAdapter
                is returned based on the git_provider field in this object

        Raises:
            GitProviderUnavailable: If the supplied git config has an unknown git provider, this error will be thrown

        Returns:
            GitAdapter: A specific subclass of the GitAdapter, based on what git_provider we need
        """
        from jf_ingest.jf_git.adapters.azure_devops import AzureDevopsAdapter
        from jf_ingest.jf_git.adapters.github import GithubAdapter

        if config.git_provider in [GitProviderInJellyfishRepo.GITHUB, GitProvider.GITHUB]:
            adapter = GithubAdapter(config)
        elif config.git_provider in [GitProviderInJellyfishRepo.ADO, GitProvider.ADO]:
            adapter = AzureDevopsAdapter(config)
        else:
            raise GitProviderUnavailable(
                f'Git provider {config.git_provider} is not currently supported'
            )

        adapter.config = config
        return adapter

    @abstractmethod
    def get_api_scopes(self) -> str:
        """Return the list of API Scopes. This is useful for Validation

        Returns:
            str: A string of API scopes we have, given the adapters credentials
        """
        pass

    @abstractmethod
    def get_organizations(self) -> List[StandardizedOrganization]:
        """Get the list of organizations the adapter has access to

        Returns:
            List[StandardizedOrganization]: A list of standardized organizations within this Git Instance
        """
        pass

    @abstractmethod
    def get_users(
        self, standardized_organization: StandardizedOrganization, limit: int = None
    ) -> Generator[StandardizedUser, None, None]:
        """Get the list of users in a given Git Organization

        Args:
            standardized_organization (StandardizedOrganization): A standardized Git Organization Object

        Returns:
            List[StandardizedUser]: A standardized User Object
            limit (int, optional): When provided, the number of items returned is limited.
                Useful for the validation use case, where we want to just verify we can pull PRs.
                Defaults to None.
        """
        pass

    @abstractmethod
    def get_teams(
        self, standardized_organization: StandardizedOrganization, limit: int = None
    ) -> Generator[StandardizedTeam, None, None]:
        """Get the list of teams in a given Git Organization

        Args:
            standardized_organization (StandardizedOrganization): A standardized Git Organization Object

        Returns:
            List[StandardizedUser]: A standardized Team Object
            limit (int, optional): When provided, the number of items returned is limited.
                Useful for the validation use case, where we want to just verify we can pull PRs.
                Defaults to None.
        """
        pass

    @abstractmethod
    def get_repos(
        self,
        standardized_organization: StandardizedOrganization,
    ) -> Generator[StandardizedRepository, None, None]:
        """Get a list of standardized repositories within a given organization

        Args:
            standardized_organization (StandardizedOrganization): A standardized organization

        Returns:
            List[StandardizedRepository]: A list of standardized Repositories
        """
        pass

    @abstractmethod
    def get_commits_for_default_branch(
        self,
        standardized_repo: StandardizedRepository,
        limit: int = None,
        pull_since: datetime = None,
        pull_until: datetime = None,
    ) -> Generator[StandardizedCommit, None, None]:
        """For a given repo, get all the commits that are on the Default Branch.

        Args:
            standardized_repo (StandardizedRepository): A standard Repository object
            limit (int): limit the number of commit objects we will yield
            pull_since (datetime): filter commits to be newer than this date
            pull_until (datetime): filter commits to be older than this date

        Returns:
            List[StandardizedCommit]: A list of standardized commits
        """
        pass

    @abstractmethod
    def get_branches_for_repo(
        self,
        standardized_repo: StandardizedRepository,
        pull_branches: Optional[bool] = False,
    ) -> Generator[StandardizedBranch, None, None]:
        """Function for pulling branches for a repository. By default, pull_branches will run as False,
        so we will only process the default branch. If pull_branches is true, than we will pull all
        branches in this repository

        Args:
            standardized_repo (StandardizedRepository): A standardized repo, which hold info about the default branch.
            pull_branches (bool): A boolean flag. If True, pull all branches available on Repo. If false, only process the default branch. Defaults to False.

        Yields:
            StandardizedBranch: A Standardized Branch Object
        """
        pass

    @abstractmethod
    def get_commits_for_branches(
        self,
        standardized_repo: StandardizedRepository,
        branches: List[StandardizedBranch],
        pull_since: datetime = None,
        pull_until: datetime = None,
    ) -> Generator[StandardizedCommit, None, None]:
        """For a given repo, get all the commits that are on the included branches.
        Included branches are found by crawling across the branches pulled/available
        from get_filtered_branches

        Args:
            standardized_repo (StandardizedRepository): A standard Repository object
            pull_since (datetime): A date to pull from
            pull_until (datetime): A date to pull up to

        Returns:
            List[StandardizedCommit]: A list of standardized commits
        """
        pass

    @abstractmethod
    def get_pr_metadata(
        self,
        standardized_repo: StandardizedRepository,
        limit: int = None,
    ) -> Generator[StandardizedPullRequestMetadata, None, None]:
        """Get all PRs, but only included the bare necesaties

        Args:
            standardized_repo (StandardizedRepository): A standardized repository
            limit (int, optional): When provided, the number of items returned is limited.
                Useful for the validation use case, where we want to just verify we can pull PRs.
                Defaults to None.

        Returns:
            List[StandardizedPullRequest]: A list of standardized PRs
        """
        pass

    @abstractmethod
    def git_provider_pr_endpoint_supports_date_filtering(self) -> bool:
        """Returns a boolean on if this PR supports time window filtering.
        So far, Github DOES NOT support this (it's adapter will return False)
        but ADO does support this (it's adapter will return True)

        Returns:
            bool: A boolean on if the adapter supports time filtering when searching for PRs
        """
        return False

    @abstractmethod
    def get_prs(
        self,
        standardized_repo: StandardizedRepository,
        pull_files_for_pr: bool = False,
        hash_files_for_prs: bool = False,
        limit: int = None,
        start_cursor: Any = None,
        start_window: datetime = None,
        end_window: datetime = None,
    ) -> Generator[StandardizedPullRequest, None, None]:
        """Get the list of standardized Pull Requests for a Standardized Repository.

        Args:
            standardized_repo (StandardizedRepository): A standardized repository
            pull_files_for_pr (bool): When provided, we will pull file metadata for all PRs
            hash_files_for_prs (bool): When provided, all file metadata will be hashed for PRs
            limit (int, optional): When provided, the number of items returned is limited.
                Useful for the validation use case, where we want to just verify we can pull PRs.
                Defaults to None.

        Returns:
            List[StandardizedPullRequest]: A list of standardized PRs
        """
        pass

    def get_commits_for_repo(
        self, standardized_repo: StandardizedRepository, branches: List[StandardizedBranch]
    ) -> Generator[StandardizedCommit, None, None]:
        """This is a function that wraps the get_commits_for_branches function and applies the Repo
        backpopulation logic, if we need it to

        Args:
            standardized_repo (StandardizedRepository): A standardized Repository object
            branches (List[StandardizedBranches]): A list of branches to pull commits for

        Yields:
            Generator[StandardizedCommit, None, None]: A stream of commits. Potentially terminating early if we hit the pull from date
        """
        pull_from_for_commits = self.config.get_pull_from_for_commits()
        backpopulation_window = determine_commit_backpopulation_window(
            config=self.config, repo=standardized_repo
        )
        pull_until_for_commits = None

        # If backpopulating, set
        if backpopulation_window:
            backpopulation_start, backpopulation_end = backpopulation_window
            pull_from_for_commits = backpopulation_start
            pull_until_for_commits = backpopulation_end
            logging_helper.send_to_agent_log_file(
                f'Backpopulation was determeined as necessary for {standardized_repo.name}. Backpopulation will run from [{pull_from_for_commits}, {pull_until_for_commits}]'
            )
        else:
            logging_helper.send_to_agent_log_file(
                f'Backpopulation was not determined as necessary for {standardized_repo.name}. Commits will be pulled from {pull_from_for_commits}'
            )

        commit_count = 0
        commit = None
        for j, commit in enumerate(
            self.get_commits_for_branches(
                standardized_repo=standardized_repo,
                branches=branches,
                pull_since=pull_from_for_commits,
                pull_until=pull_until_for_commits,
            ),
            start=1,
        ):
            with logging_helper.log_loop_iters('branch commit inside repo', j, 100):
                # If we crawl across commits and find that we already have commits this old, stop processing
                # NOTE: THis is technically redundant, because the get_commits calls should have a pull_from/pull_until
                # scheme that should limit how many commits we pull
                if commit.commit_date and commit.commit_date < pull_from_for_commits:
                    break
                yield commit
                commit_count += 1
        if backpopulation_window:
            commits_backpopulated_to = None
            if commit:
                commits_backpopulated_to = max(
                    min(pull_from_for_commits, commit.commit_date), self.config.pull_from
                )
            else:
                commits_backpopulated_to = max(pull_from_for_commits, self.config.pull_from)
            standardized_repo.commits_backpopulated_to = commits_backpopulated_to
            logging_helper.send_to_agent_log_file(
                f'Setting commits_backpopulated_to for repo {standardized_repo.name} to {commits_backpopulated_to}'
            )
        logging_helper.send_to_agent_log_file(f'Found {commit_count} commits', level=logging.DEBUG)

    def get_prs_for_repo(
        self,
        standardized_repo: StandardizedRepository,
        pull_files_for_pr: bool,
        hash_files_for_prs: bool,
    ) -> Generator[StandardizedPullRequest, None, None]:
        """This is a function that wraps the get_commits_for_branches function and applies the Repo
        backpopulation logic, if we need it to

        Args:
            standardized_repo (StandardizedRepository): A standardized Repository object

        Yields:
            Generator[StandardizedCommit, None, None]: A stream of commits. Potentially terminating early if we hit the pull from date
        """
        prs_start_cursor = None
        backpopulation_window = determine_pr_backpopulation_window(
            config=self.config, repo=standardized_repo
        )
        if backpopulation_window:
            pull_from_for_prs, pull_up_to_for_prs = backpopulation_window
        else:
            pull_from_for_prs = self.config.get_pull_from_for_prs(standardized_repo.id)
            pull_up_to_for_prs = datetime.now().astimezone(timezone.utc) + timedelta(days=1)

        # If we are backpopulating and our Adapter DOES NOT support filtering for PRs with
        # datetime bounds, we need to find the starting mark of where to start
        # pulling PRs. To do this, we leverage the get_pr_metadata function, which should be
        # a light-weight alternative (in terms of API calls) to the get_prs function.
        # For an adapter that uses GQL, this alternative can be VERY light. For a non-GQL
        # adapter, this can be slightly lighter but likely not by much
        # NOTE: If a provider supports PR time filtering (like ADO), then this can be skipped!
        # It is faster to have the API do the filtering for us
        if backpopulation_window and not self.git_provider_pr_endpoint_supports_date_filtering():
            logging_helper.send_to_agent_log_file(
                f'Backpopulation window detected for PRs in {standardized_repo.name}, attempting to walk back on all PRs to find backpopulation window end date'
            )

            backpopulation_start, backpopulation_end = backpopulation_window
            pull_from_for_prs = backpopulation_start
            prs_found = False
            for api_pr_metadata in self.get_pr_metadata(standardized_repo=standardized_repo):
                if api_pr_metadata.updated_at > backpopulation_end:
                    logging_helper.send_to_agent_log_file(
                        f'Backpopulation flow -- skipping PR (ID: {api_pr_metadata.id}) from {api_pr_metadata.updated_at} '
                        f'because backpopulation_end is {backpopulation_end.isoformat()} (Repo: {standardized_repo.name})'
                    )
                    # This is the START cursor, so it is NON-INCLUSIVE. We want it to be trailing by 1 index
                    prs_start_cursor = api_pr_metadata.api_index
                    continue
                else:
                    if api_pr_metadata.updated_at <= self.config.pull_from:
                        logging_helper.send_to_agent_log_file(
                            f'Exiting backpopulation walkback loop and NOT ingesting this PR, because PR {api_pr_metadata.id} was last updated at {api_pr_metadata.updated_at} which is less than our base pull from date: {self.config.pull_from}'
                        )
                        standardized_repo.prs_backpopulated_to = self.config.pull_from
                        return
                    elif api_pr_metadata.updated_at <= backpopulation_start:
                        # We want to ingest this one PR in this case, because it will greatly fast forward our backpopulation dates
                        logging_helper.send_to_agent_log_file(
                            f'Exiting backpopulation walkback loop, because PR {api_pr_metadata.id} was last updated at {api_pr_metadata.updated_at} which is less than our backpopulation start time ({backpopulation_start}). We WILL ingest this PR'
                        )
                        prs_found = True
                        break
                    else:
                        logging_helper.send_to_agent_log_file(
                            f'Exiting backpopulation walkback loop, because PR {api_pr_metadata.id} was last updated at {api_pr_metadata.updated_at} which is within our backpopulation window ([{backpopulation_start}, {backpopulation_end}]). We will ingest this PR and all other PRs up until {backpopulation_start}'
                        )
                        prs_found = True
                        break

            if not prs_found:
                logging_helper.send_to_agent_log_file(
                    f'No PRs found when looking in and beyond our backpopulation window, setting PRs backpopulated to to the pull from date for this git instance {self.config.pull_from}'
                )
                standardized_repo.prs_backpopulated_to = self.config.pull_from
                return

        pr = None
        pr_count_for_repo = 0
        for i, pr in enumerate(
            self.get_prs(
                standardized_repo=standardized_repo,
                pull_files_for_pr=pull_files_for_pr,
                hash_files_for_prs=hash_files_for_prs,
                start_cursor=prs_start_cursor,
                start_window=pull_from_for_prs,
                end_window=pull_up_to_for_prs,
            ),
            start=1,
        ):
            with logging_helper.log_loop_iters('pr inside repo', i, 10):
                # If we crawl across prs and find that we already have PR this old, stop processing
                if (
                    not self.git_provider_pr_endpoint_supports_date_filtering()
                    and pr.updated_at
                    and pr.updated_at <= pull_from_for_prs
                ):
                    logging_helper.send_to_agent_log_file(
                        f'Stopping PR crawl for repo {standardized_repo.name} because PR {pr.id} as been identified as being older than the pull from date ({pr.updated_at} <= {pull_from_for_prs}).'
                    )
                    # If we're backpopulating, this PR represents the next oldest PR. If we ingest it, we can speed up
                    # the backpopulation window to be as old as this PR. Only ingest it if it's within the parent 'pull_from' window, though
                    if backpopulation_window and pr.updated_at >= self.config.pull_from:
                        logging_helper.send_to_agent_log_file(
                            f'This PR ({pr.id}) will be ingest by Jellyfish, because are backpopulating this repo ({standardized_repo.name})'
                        )
                        yield pr
                        pr_count_for_repo += 1
                    break
                yield pr
                pr_count_for_repo += 1

        # If we're backpopulating, update the prs_back_populated_to variable
        if backpopulation_window:
            prs_back_populated_to = None
            if pr:
                prs_back_populated_to = max(
                    min(pull_from_for_prs, pr.updated_at), self.config.pull_from
                )
            else:
                prs_back_populated_to = max(pull_from_for_prs, self.config.pull_from)
            standardized_repo.prs_backpopulated_to = prs_back_populated_to
            logging_helper.send_to_agent_log_file(
                f'Setting prs_backpopulated_to for repo {standardized_repo.name} to {prs_back_populated_to}'
            )

        logging_helper.send_to_agent_log_file(
            f'{pr_count_for_repo} PRs found for repo {standardized_repo.name}'
        )

    def get_filtered_branches(
        self, repo: StandardizedRepository, branches: List[StandardizedBranch]
    ) -> set[str]:
        """Return branches for which we should pull commits, specified by customer in git config.
            The repo's default branch will always be included in the returned list.

        Args:
            repo (StandardizedRepository): A standardized repository

        Returns:
            set[str]: A set of branch names (as strings)
        """

        # Helper function
        def get_matching_branches(
            included_branch_patterns: List[str], repo_branch_names: List[str]
        ) -> List[str]:
            # Given a list of patterns, either literal branch names or names with wildcards (*) meant to match a set of branches in a repo,
            # return the list of branches from repo_branches that match any of the branch name patterns.
            # fnmatch is used over regex to support wildcards but avoid complicating the requirements on branch naming in a user's config.
            matching_branches = []
            for repo_branch_name in repo_branch_names:
                if self.config.pull_all_commits_and_branches:
                    matching_branches.append(repo_branch_name)
                elif any(
                    fnmatch.fnmatch(repo_branch_name, pattern)
                    for pattern in included_branch_patterns
                ):
                    matching_branches.append(repo_branch_name)
            return matching_branches

        # Always process the default branch
        branches_to_process = [repo.default_branch_name] if repo.default_branch_name else []
        # Agent use case: check for the included_branches values
        additional_branches_for_repo = self.config.included_branches_by_repo.get(repo.name, [])

        # Extend and potentially filter branches to process
        repo_branch_names = [b.name for b in branches if b]
        branches_to_process.extend(
            get_matching_branches(additional_branches_for_repo, repo_branch_names)
        )
        return set(branches_to_process)

    def discover_new_orgs(self) -> List[StandardizedOrganization]:
        """Helper function for discovering new Git organizations. Currently
        not implemented because only Github has been implement as a GitAdapter
        subclass, and Github DOES NOT support discovering new orgs. Orgs must
        be entered manually

        Raises:
            NotImplementedError: Error stating that this function is net yet to use yet

        Returns:
            List[StandardizedOrganization]: A list of standardized Org Objects
        """
        raise NotImplementedError('Discover New Orgs is not yet implemented')

    def load_and_dump_git(self, git_config: GitConfig, ingest_config: IngestionConfig):
        """This is a shared class function that can get called by
        the different types of GitAdapters that extend this class.
        This function handles fetching all the necessary data from
        Git, transforming it, and saving it to local disk and/or S3

        Args:
            ingest_config (IngestionConfig): A valid Ingestion Config
        """
        init_jf_ingest_run(ingestion_config=ingest_config)
        with emit_event_to_jellyfish_context_manager(
            _generate_git_ingest_event(
                event_name=GitIngestionEventNames.GET_GIT_DATA, git_config=git_config
            )
        ):
            self._run_load_and_dump_git(git_config=git_config, ingest_config=ingest_config)

    def _run_load_and_dump_git(self, git_config: GitConfig, ingest_config: IngestionConfig):
        #######################################################################
        # Init IO Helper
        #######################################################################
        ingest_io_helper = IngestIOHelper(ingest_config=ingest_config)

        # Wrapper function for writing to the IngestIOHelper
        def _write_to_s3_or_local(object_name: str, json_data: list[dict], batch_number: int = 0):
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=object_name,
                json_data=json_data,
                subdirectory=SubDirectory.GIT,
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
                git_instance_key=self.config.instance_file_key,
                batch_number=batch_number,
                enable_telemetry=False,
            )

        #######################################################################
        # ORGANIZATION DATA
        #######################################################################
        with emit_event_to_jellyfish_context_manager(
            _generate_git_ingest_event(
                event_name=GitIngestionEventNames.GET_GIT_ORGANIZATIONS, git_config=git_config
            )
        ):
            logger.info('Fetching Git Organization Data...')
            standardized_organizations: List[StandardizedOrganization] = self.get_organizations()
            logger.info(
                f'Successfully pulled Git Organizations data for {len(standardized_organizations)} Organizations.'
            )
            # Upload Data
            _write_to_s3_or_local(
                object_name=GitObject.GitOrganizations.value,
                json_data=_transform_dataclass_list_to_dict(standardized_organizations),
            )

        #######################################################################
        # USER DATA
        #######################################################################
        with emit_event_to_jellyfish_context_manager(
            _generate_git_ingest_event(
                event_name=GitIngestionEventNames.GET_GIT_USERS, git_config=git_config
            )
        ):
            if not git_config.skip_pulling_users:
                logger.info('Fetching Git User Data...')
                standardized_users: List[StandardizedUser] = [
                    user
                    for org in standardized_organizations
                    for user in tqdm_to_logger(
                        self.get_users(org), desc='Processing Users', unit=' users'
                    )
                ]
                logger.info(f'Successfully found {len(standardized_users)} users.')
                # Upload Data
                _write_to_s3_or_local(
                    object_name=GitObject.GitUsers.value,
                    json_data=_transform_dataclass_list_to_dict(standardized_users),
                )
            else:
                _write_to_s3_or_local(
                    object_name=GitObject.GitUsers.value,
                    json_data=[],
                )
                logger.info(
                    f'Not pulling users because \'skip_pulling_users\' is set to: {git_config.skip_pulling_users}.'
                )

        #######################################################################
        # TEAM DATA
        #######################################################################
        with emit_event_to_jellyfish_context_manager(
            _generate_git_ingest_event(
                event_name=GitIngestionEventNames.GET_GIT_TEAMS, git_config=git_config
            )
        ):
            logger.info('Fetching Git Team Data...')
            standardized_teams: List[StandardizedTeam] = (
                [
                    team
                    for org in standardized_organizations
                    for team in tqdm_to_logger(
                        self.get_teams(org), desc="Processing Teams", unit=" teams"
                    )
                ]
                if self.config.pull_teams
                else []
            )
            logger.info(f'Successfully found {len(standardized_teams)} teams.')
            # Upload Data
            _write_to_s3_or_local(
                object_name=GitObject.GitTeams.value,
                json_data=_transform_dataclass_list_to_dict(standardized_teams),
            )

        #######################################################################
        # REPO DATA, NOTE THAT WE UPLOAD LATER BECAUSE WE NEED TO SET
        # THE BACKPOPULATD DATES BELOW AFTER WE PULL PRS AND COMMITS
        #######################################################################
        if not git_config.skip_pulling_repos:
            with emit_event_to_jellyfish_context_manager(
                _generate_git_ingest_event(
                    event_name=GitIngestionEventNames.GET_GIT_REPOS, git_config=git_config
                )
            ):
                logger.info('Fetching Git Repo Data...')
                standardized_repos: List[StandardizedRepository] = [
                    repo
                    for org in standardized_organizations
                    for repo in tqdm_to_logger(
                        self.get_repos(
                            standardized_organization=org,
                        ),
                        unit=' repos',
                        desc=f'Pulling all available Repositories',
                    )
                ]
                logger.info(
                    f'Successfully pulled Git Repo Data for {len(standardized_repos)} Repos.'
                )
        else:
            logger.info(
                f'Not pulling new repo data because \'skip_pulling_repos\' is set to {git_config.skip_pulling_repos}. '
                f'We will pull data for the {len(git_config.repos_in_jellyfish)} repos that already exist in Jellyfish'
            )
            standardized_repos = git_config.repos_in_jellyfish

        repos_to_process = [
            repo for repo in standardized_repos if repo.id not in git_config.quiescent_repos
        ]

        filters = []
        if self.config.included_repos:
            logger.info(f'Filtering repos to only include {self.config.included_repos}')
            filters.append(
                lambda repo_name: repo_name.lower()
                in set([r.lower() for r in self.config.included_repos])
            )
        if self.config.excluded_repos:
            logger.info(f'Filtering repos to exclude {self.config.excluded_repos}')
            filters.append(
                lambda repo_name: repo_name.lower()
                not in set([r.lower() for r in self.config.excluded_repos])
            )

        repos_to_process = [
            repo for repo in repos_to_process if all(filt(repo.name) for filt in filters)
        ]

        repo_count = len(repos_to_process)

        logging_helper.send_to_agent_log_file(
            f'Processing {len(repos_to_process)}. {len(git_config.quiescent_repos)} were marked as being quiescent'
        )

        #######################################################################
        # BRANCH DATA
        # NOTE: Branches are optionally processed, depending on GitConfiguration.
        # For Direct Connect it is likely we only process the default branch,
        # for agent we process all branches
        #######################################################################
        repo_to_branches: dict[str, List[StandardizedBranch]] = {}
        all_branches = []
        with tqdm_to_logger(
            desc='Processing Branches', unit=' Branches'
        ) as pbar, emit_event_to_jellyfish_context_manager(
            _generate_git_ingest_event(
                event_name=GitIngestionEventNames.GET_GIT_BRANCHES, git_config=git_config
            )
        ):
            for repo in repos_to_process:
                pull_branches = (
                    git_config.pull_all_commits_and_branches
                    or git_config.repo_id_to_pull_all_commits_and_branches.get(repo.id)
                )
                branch_batch = []
                # Iterate across branches and update the progress bar so we can see
                # counts and rates of branch processing
                for branch in self.get_branches_for_repo(repo, pull_branches):
                    pbar.update(1)
                    branch_batch.append(branch)

                repo_to_branches[repo.id] = branch_batch
                all_branches.extend(branch_batch)

        _write_to_s3_or_local(
            object_name=GitObject.GitBranches.value,
            json_data=_transform_dataclass_list_to_dict(all_branches),
        )

        #######################################################################
        # COMMIT DATA
        #
        # NOTE: Commit data can be quite large, so for better memory handling
        # we will create a chain of generators (get_commits_for_branches returns a generator)
        # and process our way through those generators, uploading data ~50 MBs at a time
        # NOTE: Commit data is pretty uniform in size (each commit is ~2KB), so we'll upload
        # in batches of 30k commits (roughly 50 MB in data)
        #
        #######################################################################
        with emit_event_to_jellyfish_context_manager(
            _generate_git_ingest_event(
                event_name=GitIngestionEventNames.GET_GIT_COMMITS, git_config=git_config
            )
        ):
            total_commits = 0
            logger.info(f'Fetching Git Commit Data for {repo_count} Repos...')
            list_of_commit_generators: List[Generator[StandardizedCommit, None, None]] = []
            for repo in repos_to_process:
                branches = repo_to_branches[repo.id]
                commit_generator_for_repo = self.get_commits_for_repo(repo, branches=branches)
                transformed_commit_generator = map(lambda commit: commit, commit_generator_for_repo)
                list_of_commit_generators.append(transformed_commit_generator)

            # Chain together all the generators
            commits_generator = tqdm_to_logger(
                chain.from_iterable(list_of_commit_generators),
                desc=f'Processing Commits for {repo_count} repos',
                unit=' commits',
            )
            for batch_num, commit_batch in enumerate(
                batch_iterable(commits_generator, batch_size=self.NUMBER_OF_COMMITS_PER_BATCH)
            ):
                total_commits += len(commit_batch)
                commit_batch_as_dict = _transform_dataclass_list_to_dict(commit_batch)
                _write_to_s3_or_local(
                    object_name=GitObject.GitCommits.value,
                    json_data=commit_batch_as_dict,
                    batch_number=batch_num,
                )
            if not total_commits:
                _write_to_s3_or_local(object_name=GitObject.GitCommits.value, json_data=[])
            logger.info(f'Successfully process {total_commits} total commits')

        #######################################################################
        # PULL REQUEST DATA
        #
        # NOTE: Pull Request data can be quite large, so for better memory handling
        # we will create a chain of generators (get_prs returns a generator)
        # and process our way through those generators, uploading data ~50 MBs at a time
        #
        #######################################################################
        with emit_event_to_jellyfish_context_manager(
            _generate_git_ingest_event(
                event_name=GitIngestionEventNames.GET_GIT_PULL_REQUESTS, git_config=git_config
            )
        ):
            total_prs = 0
            logger.info(f'Fetching Git Pull Request Data for {repo_count} Repos...')
            list_of_pr_generators: List[Generator[StandardizedPullRequest, None, None]] = []
            for repo in repos_to_process:
                if self.config.repos_to_skip_pull_prs_for.get(repo.id):
                    logging_helper.send_to_agent_log_file(
                        f'Skipping pull PRs for {repo.name} because it was marked in Jellyfish as so'
                    )
                    continue
                pr_generator_for_repo = self.get_prs_for_repo(
                    repo,
                    pull_files_for_pr=git_config.pull_files_for_prs,
                    hash_files_for_prs=git_config.hash_files_for_prs,
                )
                transformed_pr_generator = map(lambda pr: asdict(pr), pr_generator_for_repo)
                list_of_pr_generators.append(transformed_pr_generator)

            # Chain together all the generators
            prs_generator = tqdm_to_logger(
                chain.from_iterable(list_of_pr_generators),
                desc=f'Processing Pull Request Data for {repo_count} repos',
                unit=' PRs',
            )
            for batch_num, pr_batch in enumerate(
                batch_iterable_by_bytes_size(
                    prs_generator, batch_byte_size=self.PULL_REQUEST_BATCH_SIZE_IN_BYTES
                )
            ):
                total_prs += len(pr_batch)
                _write_to_s3_or_local(
                    object_name=GitObject.GitPullRequests.value,
                    json_data=pr_batch,
                    batch_number=batch_num,
                )

            if not total_prs:
                # IF we don't have any PRs, push an empty file
                _write_to_s3_or_local(
                    object_name=GitObject.GitPullRequests.value,
                    json_data=[],
                    batch_number=0,
                )

            logger.info(f'Successfully processed {total_prs} total PRs')

        # Upload Repo Data at the very end
        _write_to_s3_or_local(
            object_name=GitObject.GitRepositories.value,
            json_data=_transform_dataclass_list_to_dict(standardized_repos)
            if standardized_repos
            else [],
        )

        logger.info(f'Done processing Git Data!')


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def load_and_push_git_to_s3(ingest_config: IngestionConfig):
    """Handler function for the end to end processing of Git Data.
    This function is responsible for taking in an ingest config,
    creating a git adapter, and then running the Git Adapter function
    for uploading data to S3 (or saving it locally). The function for
    handling that logic is part of the GitAdapter class (see load_and_dump_git)

    Args:
        ingest_config (IngestionConfig): A fully formed IngestionConfig class, with
        valid Git Configuration in it.
    """
    for git_config in ingest_config.git_configs:
        try:
            git_adapter: GitAdapter = GitAdapter.get_git_adapter(git_config)
            git_adapter.load_and_dump_git(git_config=git_config, ingest_config=ingest_config)
        except GitProviderUnavailable:
            logger.warning(
                f'Git Config for provider {git_config.git_provider} is currently NOT supported!'
            )
            continue


def determine_commit_backpopulation_window(
    config: GitConfig, repo: StandardizedRepository
) -> BackpopulationWindow:
    """Get the backpopulation window for Commits

    Args:
        config (GitConfig): A valid Git Config
        repo (StandardizedRepository): A valid standardized repository

    Returns:
        BackpopulationWindow: A Backpopulation window object
    """
    commits_backpopulated_to = config.get_backpopulated_date_for_commits(repo.id)
    return _get_backpopulation_helper(
        repo=repo,
        pull_from=config.pull_from,
        objects_back_populated_to=commits_backpopulated_to,
        object_name='commits',
        force_full_backpopulation_pull=config.force_full_backpopulation_pull,
        backpopulation_window_days=config.backpopulation_window_days,
    )


def determine_pr_backpopulation_window(
    config: GitConfig, repo: StandardizedRepository
) -> BackpopulationWindow:
    """Get the backpopulation window for PRs

    Args:
        config (GitConfig): A valid Git Config
        repo (StandardizedRepository): A valid standardized repository

    Returns:
        BackpopulationWindow: A Backpopulation window object
    """
    prs_backpopulated_to = config.get_backpopulated_date_for_prs(repo.id)
    return _get_backpopulation_helper(
        repo=repo,
        pull_from=config.pull_from,
        objects_back_populated_to=prs_backpopulated_to,
        object_name='PRs',
        force_full_backpopulation_pull=config.force_full_backpopulation_pull,
        backpopulation_window_days=config.backpopulation_window_days,
    )


def _get_backpopulation_helper(
    repo: StandardizedRepository,
    pull_from: date,
    objects_back_populated_to: date,
    object_name: str,
    force_full_backpopulation_pull: bool = False,
    backpopulation_window_days: int = 30,
) -> BackpopulationWindow:
    if objects_back_populated_to and objects_back_populated_to <= pull_from:
        # No backpopulation necessary
        return None
    # We're backpopulating objects for this repo

    if objects_back_populated_to:
        base_date = objects_back_populated_to
    else:
        base_date = datetime.now().astimezone(timezone.utc) + timedelta(days=1)

    backpopulation_window_start = (
        pull_from
        if force_full_backpopulation_pull
        else max(pull_from, base_date - timedelta(days=backpopulation_window_days))
    )
    backpopulation_window_end = base_date

    logging_helper.send_to_agent_log_file(
        f'Backpopulation window found for {object_name} for repo {repo.name} (ID: {repo.id}). Window spans from {backpopulation_window_start} to {backpopulation_window_end} ({object_name} backpopulated to {objects_back_populated_to}, pull_from: {pull_from})'
    )
    return BackpopulationWindow(backpopulation_window_start, backpopulation_window_end)
